# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import asyncio
import json
import threading
import time
from unittest.mock import MagicMock, Mock, patch

import pytest
from common_test_utils import MockS3
from jsonschema import validate

from braket.ahs.analog_hamiltonian_simulation import AnalogHamiltonianSimulation
from braket.annealing.problem import Problem, ProblemType
from braket.aws import AwsQuantumTask
from braket.aws.aws_quantum_task import _create_annealing_device_params
from braket.aws.aws_session import AwsSession
from braket.aws.queue_information import QuantumTaskQueueInfo, QueueType
from braket.circuits import Circuit
from braket.circuits.gates import PulseGate
from braket.circuits.serialization import (
    IRType,
    OpenQASMSerializationProperties,
    QubitReferenceType,
)
from braket.device_schema import GateModelParameters, error_mitigation
from braket.device_schema.dwave import (
    Dwave2000QDeviceParameters,
    DwaveAdvantageDeviceParameters,
    DwaveDeviceParameters,
)
from braket.device_schema.ionq import IonqDeviceParameters
from braket.device_schema.oqc import OqcDeviceParameters
from braket.device_schema.rigetti import RigettiDeviceParameters
from braket.device_schema.simulators import GateModelSimulatorDeviceParameters
from braket.error_mitigation.debias import Debias
from braket.ir.blackbird import Program as BlackbirdProgram
from braket.ir.openqasm import Program as OpenQASMProgram
from braket.pulse import Frame, Port, PulseSequence
from braket.tasks import (
    AnalogHamiltonianSimulationQuantumTaskResult,
    AnnealingQuantumTaskResult,
    GateModelQuantumTaskResult,
    PhotonicModelQuantumTaskResult,
)

S3_TARGET = AwsSession.S3DestinationFolder("foo", "bar")

IONQ_ARN = "device/qpu/ionq"
RIGETTI_ARN = "device/qpu/rigetti"
OQC_ARN = "device/qpu/oqc"
SIMULATOR_ARN = "device/quantum-simulator"
XANADU_ARN = "device/qpu/xanadu"

DEVICE_PARAMETERS = [
    (IONQ_ARN, IonqDeviceParameters),
    (RIGETTI_ARN, RigettiDeviceParameters),
    (OQC_ARN, OqcDeviceParameters),
    (SIMULATOR_ARN, GateModelSimulatorDeviceParameters),
]


@pytest.fixture
def aws_session():
    mock = Mock()
    _mock_metadata(mock, "RUNNING")
    return mock


@pytest.fixture
def quantum_task(aws_session):
    return AwsQuantumTask("foo:bar:arn", aws_session, poll_timeout_seconds=2)


@pytest.fixture
def quantum_task_quiet(aws_session):
    return AwsQuantumTask("foo:bar:arn", aws_session, poll_timeout_seconds=2, quiet=True)


@pytest.fixture
def circuit_task(aws_session):
    return AwsQuantumTask("foo:bar:arn", aws_session, poll_timeout_seconds=2)


@pytest.fixture
def annealing_task(aws_session):
    return AwsQuantumTask("foo:bar:arn", aws_session, poll_timeout_seconds=2)


@pytest.fixture
def photonic_model_task(aws_session):
    return AwsQuantumTask("foo:bar:arn", aws_session, poll_timeout_seconds=2)


@pytest.fixture
def arn():
    return "foo:bar:arn"


@pytest.fixture
def circuit():
    return Circuit().h(0).cnot(0, 1)


@pytest.fixture
def problem():
    return Problem(ProblemType.ISING, linear={1: 3.14}, quadratic={(1, 2): 10.08})


@pytest.fixture
def openqasm_program():
    return OpenQASMProgram(source="OPENQASM 3.0; h $0;")


@pytest.fixture
def blackbird_program():
    return BlackbirdProgram(source="Vac | q[0]")


@pytest.fixture
def pulse_sequence():
    return PulseSequence().set_frequency(
        Frame(
            frame_id="predefined_frame_1",
            frequency=2e9,
            port=Port(port_id="device_port_x0", dt=1e-9, properties={}),
            phase=0,
            is_predefined=True,
        ),
        6e6,
    )


@pytest.fixture
def pulse_gate(pulse_sequence):
    return PulseGate(pulse_sequence, 1, "my_PG")


@pytest.fixture
def em():
    return Debias()


@pytest.fixture
def ahs_problem():
    mock = Mock(spec=AnalogHamiltonianSimulation)
    mock.to_ir.return_value.json.return_value = "Test AHS Problem"
    return mock


def test_equality(arn, aws_session):
    quantum_task_1 = AwsQuantumTask(arn, aws_session)
    quantum_task_2 = AwsQuantumTask(arn, aws_session)
    other_quantum_task = AwsQuantumTask("different:arn", aws_session)
    non_quantum_task = quantum_task_1.id

    assert quantum_task_1 == quantum_task_2
    assert quantum_task_1 is not quantum_task_2
    assert quantum_task_1 != other_quantum_task
    assert quantum_task_1 != non_quantum_task


def test_str(quantum_task):
    expected = "AwsQuantumTask('id/taskArn':'{}')".format(quantum_task.id)
    assert str(quantum_task) == expected


def test_hash(quantum_task):
    assert hash(quantum_task) == hash(quantum_task.id)


def test_id_getter(arn, aws_session):
    quantum_task = AwsQuantumTask(arn, aws_session)
    assert quantum_task.id == arn


@pytest.mark.xfail(raises=AttributeError)
def test_no_id_setter(quantum_task):
    quantum_task.id = 123


def test_metadata(quantum_task):
    metadata_1 = {"status": "RUNNING"}
    quantum_task._aws_session.get_quantum_task.return_value = metadata_1
    assert quantum_task.metadata() == metadata_1
    quantum_task._aws_session.get_quantum_task.assert_called_with(quantum_task.id)

    metadata_2 = {"status": "COMPLETED"}
    quantum_task._aws_session.get_quantum_task.return_value = metadata_2
    assert quantum_task.metadata(use_cached_value=True) == metadata_1


def test_metadata_call_if_none(quantum_task):
    metadata_1 = {"status": "RUNNING"}
    quantum_task._aws_session.get_quantum_task.return_value = metadata_1
    assert quantum_task.metadata(use_cached_value=True) == metadata_1
    quantum_task._aws_session.get_quantum_task.assert_called_with(quantum_task.id)


def test_has_reservation_arn_from_metadata(quantum_task):
    metadata_true = {
        "associations": [
            {
                "arn": "123",
                "type": "RESERVATION_TIME_WINDOW_ARN",
            }
        ]
    }
    assert quantum_task._has_reservation_arn_from_metadata(metadata_true)

    metadata_false = {
        "status": "RUNNING",
        "associations": [
            {
                "arn": "123",
                "type": "other",
            }
        ],
    }
    assert not quantum_task._has_reservation_arn_from_metadata(metadata_false)


def test_queue_position(quantum_task):
    state_1 = "QUEUED"
    _mock_metadata(quantum_task._aws_session, state_1)
    assert quantum_task.queue_position() == QuantumTaskQueueInfo(
        queue_type=QueueType.NORMAL, queue_position="2", message=None
    )

    state_2 = "COMPLETED"
    message = (
        f"'Task is in {state_2} status. AmazonBraket does not show queue position for this status.'"
    )
    _mock_metadata(quantum_task._aws_session, state_2)
    assert quantum_task.queue_position() == QuantumTaskQueueInfo(
        queue_type=QueueType.NORMAL, queue_position=None, message=message
    )


def test_queued_quiet(quantum_task_quiet):
    state_1 = "QUEUED"
    _mock_metadata(quantum_task_quiet._aws_session, state_1)
    assert quantum_task_quiet.queue_position() == QuantumTaskQueueInfo(
        queue_type=QueueType.NORMAL, queue_position="2", message=None
    )

    state_2 = "COMPLETED"
    message = (
        f"'Task is in {state_2} status. AmazonBraket does not show queue position for this status.'"
    )
    _mock_metadata(quantum_task_quiet._aws_session, state_2)
    assert quantum_task_quiet.queue_position() == QuantumTaskQueueInfo(
        queue_type=QueueType.NORMAL, queue_position=None, message=message
    )


def test_state(quantum_task):
    state_1 = "RUNNING"
    _mock_metadata(quantum_task._aws_session, state_1)
    assert quantum_task.state() == state_1
    quantum_task._aws_session.get_quantum_task.assert_called_with(quantum_task.id)

    state_2 = "COMPLETED"
    _mock_metadata(quantum_task._aws_session, state_2)
    assert quantum_task.state(use_cached_value=True) == state_1

    state_3 = "FAILED"
    _mock_metadata(quantum_task._aws_session, state_3)
    assert quantum_task.state() == state_3

    state_4 = "CANCELLED"
    _mock_metadata(quantum_task._aws_session, state_4)
    assert quantum_task.state() == state_4


def test_cancel(quantum_task):
    future = quantum_task.async_result()

    assert not future.done()
    quantum_task.cancel()

    assert quantum_task.result() is None
    assert future.cancelled()
    quantum_task._aws_session.cancel_quantum_task.assert_called_with(quantum_task.id)


def test_cancel_without_fetching_result(quantum_task):
    quantum_task.cancel()

    assert quantum_task.result() is None
    assert quantum_task._future.cancelled()
    quantum_task._aws_session.cancel_quantum_task.assert_called_with(quantum_task.id)


def asyncio_get_event_loop_side_effect(*args, **kwargs):
    yield ValueError("unit-test-exception")
    mock = MagicMock()
    while True:
        yield mock


@patch("braket.aws.aws_quantum_task.asyncio")
def test_initialize_asyncio_event_loop_if_required(mock_asyncio, quantum_task):
    mock_asyncio.get_event_loop.side_effect = asyncio_get_event_loop_side_effect()
    mock_asyncio.set_event_loop.return_value = MagicMock()
    mock_asyncio.new_event_loop.return_value = MagicMock()

    quantum_task._get_future()

    assert mock_asyncio.get_event_loop.call_count == 2
    assert mock_asyncio.set_event_loop.call_count == 1
    assert mock_asyncio.new_event_loop.call_count == 1


def test_result_circuit(circuit_task):
    _mock_metadata(circuit_task._aws_session, "COMPLETED")
    _mock_s3(circuit_task._aws_session, MockS3.MOCK_S3_RESULT_GATE_MODEL)

    expected = GateModelQuantumTaskResult.from_string(MockS3.MOCK_S3_RESULT_GATE_MODEL)
    assert circuit_task.result() == expected

    s3_bucket = circuit_task.metadata()["outputS3Bucket"]
    s3_object_key = circuit_task.metadata()["outputS3Directory"]
    circuit_task._aws_session.retrieve_s3_object_body.assert_called_with(
        s3_bucket, f"{s3_object_key}/results.json"
    )


def test_result_annealing(annealing_task):
    _mock_metadata(annealing_task._aws_session, "COMPLETED")
    _mock_s3(annealing_task._aws_session, MockS3.MOCK_S3_RESULT_ANNEALING)

    expected = AnnealingQuantumTaskResult.from_string(MockS3.MOCK_S3_RESULT_ANNEALING)
    assert annealing_task.result() == expected

    s3_bucket = annealing_task.metadata()["outputS3Bucket"]
    s3_object_key = annealing_task.metadata()["outputS3Directory"]
    annealing_task._aws_session.retrieve_s3_object_body.assert_called_with(
        s3_bucket, f"{s3_object_key}/results.json"
    )


def test_result_photonic_model(photonic_model_task):
    _mock_metadata(photonic_model_task._aws_session, "COMPLETED")
    _mock_s3(photonic_model_task._aws_session, MockS3.MOCK_S3_RESULT_PHOTONIC_MODEL)

    expected = PhotonicModelQuantumTaskResult.from_string(MockS3.MOCK_S3_RESULT_PHOTONIC_MODEL)
    assert photonic_model_task.result() == expected

    s3_bucket = photonic_model_task.metadata()["outputS3Bucket"]
    s3_object_key = photonic_model_task.metadata()["outputS3Directory"]
    photonic_model_task._aws_session.retrieve_s3_object_body.assert_called_with(
        s3_bucket, f"{s3_object_key}/results.json"
    )


def test_result_analog_hamiltonian_simulation(quantum_task):
    _mock_metadata(quantum_task._aws_session, "COMPLETED")
    _mock_s3(quantum_task._aws_session, MockS3.MOCK_S3_RESULT_ANALOG_HAMILTONIAN_SIMULTION)

    expected = AnalogHamiltonianSimulationQuantumTaskResult.from_string(
        MockS3.MOCK_S3_RESULT_ANALOG_HAMILTONIAN_SIMULTION
    )
    assert quantum_task.result() == expected

    s3_bucket = quantum_task.metadata()["outputS3Bucket"]
    s3_object_key = quantum_task.metadata()["outputS3Directory"]
    quantum_task._aws_session.retrieve_s3_object_body.assert_called_with(
        s3_bucket, f"{s3_object_key}/results.json"
    )


@pytest.mark.xfail(raises=TypeError)
def test_result_invalid_type(circuit_task):
    _mock_metadata(circuit_task._aws_session, "COMPLETED")
    _mock_s3(circuit_task._aws_session, json.dumps(MockS3.MOCK_TASK_METADATA))
    circuit_task.result()


def test_result_circuit_cached(circuit_task):
    _mock_metadata(circuit_task._aws_session, "COMPLETED")
    expected = GateModelQuantumTaskResult.from_string(MockS3.MOCK_S3_RESULT_GATE_MODEL)
    circuit_task._result = expected
    assert circuit_task.result() == expected
    assert not circuit_task._aws_session.retrieve_s3_object_body.called


def test_no_result(circuit_task):
    _mock_metadata(circuit_task._aws_session, "FAILED")
    circuit_task._result = None
    assert circuit_task.result() is None
    assert not circuit_task._aws_session.retrieve_s3_object_body.called


@pytest.mark.parametrize(
    "result_string",
    [MockS3.MOCK_S3_RESULT_GATE_MODEL, MockS3.MOCK_S3_RESULT_GATE_MODEL_WITH_RESULT_TYPES],
)
def test_result_cached_future(circuit_task, result_string):
    _mock_metadata(circuit_task._aws_session, "COMPLETED")
    _mock_s3(circuit_task._aws_session, result_string)
    circuit_task.result()

    _mock_s3(circuit_task._aws_session, "")
    expected = GateModelQuantumTaskResult.from_string(MockS3.MOCK_S3_RESULT_GATE_MODEL)
    assert circuit_task.result() == expected


@pytest.mark.parametrize(
    "status, result",
    [
        ("COMPLETED", GateModelQuantumTaskResult.from_string(MockS3.MOCK_S3_RESULT_GATE_MODEL)),
        ("FAILED", None),
    ],
)
def test_async_result(circuit_task, status, result):
    def set_result_from_callback(future):
        # Set the result_from_callback variable in the enclosing functions scope
        nonlocal result_from_callback
        result_from_callback = future.result()

    _mock_metadata(circuit_task._aws_session, "RUNNING")
    _mock_s3(circuit_task._aws_session, MockS3.MOCK_S3_RESULT_GATE_MODEL)

    future = circuit_task.async_result()

    # test the different ways to get the result from async

    # via callback
    result_from_callback = None
    future.add_done_callback(set_result_from_callback)

    # via asyncio waiting for result
    _mock_metadata(circuit_task._aws_session, status)
    event_loop = asyncio.get_event_loop()
    result_from_waiting = event_loop.run_until_complete(future)

    # via future.result(). Note that this would fail if the future is not complete.
    result_from_future = future.result()

    assert result_from_callback == result
    assert result_from_waiting == result
    assert result_from_future == result


@pytest.mark.parametrize(
    "status, result",
    [
        ("COMPLETED", GateModelQuantumTaskResult.from_string(MockS3.MOCK_S3_RESULT_GATE_MODEL)),
        ("FAILED", None),
    ],
)
def test_async_result_queued(circuit_task, status, result):
    def set_result_from_callback(future):
        # Set the result_from_callback variable in the enclosing functions scope
        nonlocal result_from_callback
        result_from_callback = future.result()

    _mock_metadata(circuit_task._aws_session, "QUEUED")
    _mock_s3(circuit_task._aws_session, MockS3.MOCK_S3_RESULT_GATE_MODEL)

    future = circuit_task.async_result()

    # test the different ways to get the result from async

    # via callback
    result_from_callback = None
    future.add_done_callback(set_result_from_callback)

    # via asyncio waiting for result
    _mock_metadata(circuit_task._aws_session, status)
    event_loop = asyncio.get_event_loop()
    result_from_waiting = event_loop.run_until_complete(future)

    # via future.result(). Note that this would fail if the future is not complete.
    result_from_future = future.result()

    assert result_from_callback == result
    assert result_from_waiting == result
    assert result_from_future == result


def test_failed_task(quantum_task):
    _mock_metadata(quantum_task._aws_session, "FAILED")
    _mock_s3(quantum_task._aws_session, MockS3.MOCK_S3_RESULT_GATE_MODEL)
    result = quantum_task.result()
    assert result is None


def test_timeout_completed(aws_session):
    _mock_metadata(aws_session, "RUNNING")
    _mock_s3(aws_session, MockS3.MOCK_S3_RESULT_GATE_MODEL)

    # Setup the poll timing such that the timeout will occur after one API poll
    quantum_task = AwsQuantumTask(
        "foo:bar:arn",
        aws_session,
        poll_timeout_seconds=0.5,
        poll_interval_seconds=1,
    )
    assert quantum_task.result() is None
    _mock_metadata(aws_session, "COMPLETED")
    assert quantum_task.state() == "COMPLETED"
    assert quantum_task.result() == GateModelQuantumTaskResult.from_string(
        MockS3.MOCK_S3_RESULT_GATE_MODEL
    )
    # Cached status is still COMPLETED, so result should be fetched
    _mock_metadata(aws_session, "RUNNING")
    quantum_task._result = None
    assert quantum_task.result() == GateModelQuantumTaskResult.from_string(
        MockS3.MOCK_S3_RESULT_GATE_MODEL
    )


def test_timeout_no_result_terminal_state(aws_session):
    _mock_metadata(aws_session, "RUNNING")
    _mock_s3(aws_session, MockS3.MOCK_S3_RESULT_GATE_MODEL)

    # Setup the poll timing such that the timeout will occur after one API poll
    quantum_task = AwsQuantumTask(
        "foo:bar:arn",
        aws_session,
        poll_timeout_seconds=0.5,
        poll_interval_seconds=1,
    )
    assert quantum_task.result() is None

    _mock_metadata(aws_session, "FAILED")
    assert quantum_task.result() is None


@pytest.mark.xfail(raises=ValueError)
def test_create_invalid_s3_folder(aws_session, arn, circuit):
    AwsQuantumTask.create(aws_session, arn, circuit, ("bucket",), 1000)


@pytest.mark.xfail(raises=TypeError)
def test_create_invalid_task_specification(aws_session, arn):
    mocked_task_arn = "task-arn-1"
    aws_session.create_quantum_task.return_value = mocked_task_arn
    AwsQuantumTask.create(aws_session, arn, "foo", S3_TARGET, 1000)


def test_create_openqasm_program(aws_session, arn, openqasm_program):
    aws_session.create_quantum_task.return_value = arn
    shots = 21
    AwsQuantumTask.create(aws_session, SIMULATOR_ARN, openqasm_program, S3_TARGET, shots)

    _assert_create_quantum_task_called_with(
        aws_session,
        SIMULATOR_ARN,
        openqasm_program.json(),
        S3_TARGET,
        shots,
    )


def test_create_openqasm_program_em(aws_session, arn, openqasm_program, em):
    aws_session.create_quantum_task.return_value = arn
    shots = 21
    AwsQuantumTask.create(
        aws_session,
        IONQ_ARN,
        openqasm_program,
        S3_TARGET,
        shots,
        device_parameters={"errorMitigation": em},
    )
    _assert_create_quantum_task_called_with(
        aws_session,
        IONQ_ARN,
        openqasm_program.json(),
        S3_TARGET,
        shots,
        device_parameters=IonqDeviceParameters(
            paradigmParameters=GateModelParameters(qubitCount=0),
            errorMitigation=[error_mitigation.Debias()],
        ),
    )


def test_create_openqasm_program_em_serialized(aws_session, arn, openqasm_program, em):
    aws_session.create_quantum_task.return_value = arn
    shots = 21
    AwsQuantumTask.create(
        aws_session,
        IONQ_ARN,
        openqasm_program,
        S3_TARGET,
        shots,
        device_parameters={"errorMitigation": em.serialize()},
    )
    _assert_create_quantum_task_called_with(
        aws_session,
        IONQ_ARN,
        openqasm_program.json(),
        S3_TARGET,
        shots,
        device_parameters=IonqDeviceParameters(
            paradigmParameters=GateModelParameters(qubitCount=0),
            errorMitigation=[error_mitigation.Debias()],
        ),
    )


def test_create_blackbird_program(aws_session, arn, blackbird_program):
    aws_session.create_quantum_task.return_value = arn
    shots = 21
    AwsQuantumTask.create(aws_session, XANADU_ARN, blackbird_program, S3_TARGET, shots)

    _assert_create_quantum_task_called_with(
        aws_session,
        XANADU_ARN,
        blackbird_program.json(),
        S3_TARGET,
        shots,
    )


def test_create_ahs_problem(aws_session, arn, ahs_problem):
    aws_session.create_quantum_task.return_value = arn
    shots = 21
    AwsQuantumTask.create(aws_session, SIMULATOR_ARN, ahs_problem, S3_TARGET, shots)

    _assert_create_quantum_task_called_with(
        aws_session,
        SIMULATOR_ARN,
        ahs_problem.to_ir().json(),
        S3_TARGET,
        shots,
    )


def test_create_task_with_reservation_arn(aws_session, arn, ahs_problem):
    aws_session.create_quantum_task.return_value = arn
    shots = 21
    reservation_arn = (
        "arn:aws:braket:us-west-2:123456789123:reservation/a1b123cd-45e6-789f-gh01-i234567jk8l9"
    )
    AwsQuantumTask.create(
        aws_session,
        SIMULATOR_ARN,
        ahs_problem,
        S3_TARGET,
        shots,
        reservation_arn=reservation_arn,
    )

    _assert_create_quantum_task_called_with(
        aws_session,
        SIMULATOR_ARN,
        ahs_problem.to_ir().json(),
        S3_TARGET,
        shots,
        reservation_arn=reservation_arn,
    )


def test_create_pulse_sequence(aws_session, arn, pulse_sequence):
    expected_openqasm = "\n".join(
        [
            "OPENQASM 3.0;",
            "cal {",
            "    set_frequency(predefined_frame_1, 6000000.0);",
            "}",
        ]
    )
    expected_program = OpenQASMProgram(source=expected_openqasm, inputs={})

    aws_session.create_quantum_task.return_value = arn
    AwsQuantumTask.create(aws_session, SIMULATOR_ARN, pulse_sequence, S3_TARGET, 10)

    _assert_create_quantum_task_called_with(
        aws_session,
        SIMULATOR_ARN,
        expected_program.json(),
        S3_TARGET,
        10,
    )


@pytest.mark.parametrize("device_arn,device_parameters_class", DEVICE_PARAMETERS)
def test_create_pulse_gate_circuit(
    aws_session, arn, pulse_sequence, device_arn, device_parameters_class
):
    pulse_gate_circuit = Circuit().pulse_gate([0, 1], pulse_sequence, "my_PG")
    expected_openqasm = "\n".join(
        (
            "OPENQASM 3.0;",
            "bit[2] b;",
            "cal {",
            "    set_frequency(predefined_frame_1, 6000000.0);",
            "}",
            "b[0] = measure $0;",
            "b[1] = measure $1;",
        )
    )

    expected_program = OpenQASMProgram(source=expected_openqasm, inputs={})

    aws_session.create_quantum_task.return_value = arn
    AwsQuantumTask.create(aws_session, device_arn, pulse_gate_circuit, S3_TARGET, 10)

    _assert_create_quantum_task_called_with(
        aws_session,
        device_arn,
        expected_program.json(),
        S3_TARGET,
        10,
        device_parameters_class(
            paradigmParameters=GateModelParameters(
                qubitCount=pulse_gate_circuit.qubit_count, disableQubitRewiring=False
            )
        ),
    )


@pytest.mark.parametrize("device_arn,device_parameters_class", DEVICE_PARAMETERS)
def test_create_circuit_with_shots(device_arn, device_parameters_class, aws_session, circuit):
    mocked_task_arn = "task-arn-1"
    aws_session.create_quantum_task.return_value = mocked_task_arn
    shots = 53

    task = AwsQuantumTask.create(aws_session, device_arn, circuit, S3_TARGET, shots)
    assert task == AwsQuantumTask(mocked_task_arn, aws_session)
    program = circuit.to_ir(ir_type=IRType.OPENQASM)
    assert program.inputs == {}

    _assert_create_quantum_task_called_with(
        aws_session,
        device_arn,
        program.json(),
        S3_TARGET,
        shots,
        device_parameters_class(
            paradigmParameters=GateModelParameters(
                qubitCount=circuit.qubit_count, disableQubitRewiring=False
            )
        ),
    )


def test_create_circuit_em(aws_session, circuit, em):
    mocked_task_arn = "task-arn-1"
    aws_session.create_quantum_task.return_value = mocked_task_arn
    shots = 53

    task = AwsQuantumTask.create(
        aws_session, IONQ_ARN, circuit, S3_TARGET, shots, device_parameters={"errorMitigation": em}
    )
    assert task == AwsQuantumTask(mocked_task_arn, aws_session)
    program = circuit.to_ir(ir_type=IRType.OPENQASM)
    assert program.inputs == {}

    _assert_create_quantum_task_called_with(
        aws_session,
        IONQ_ARN,
        program.json(),
        S3_TARGET,
        shots,
        IonqDeviceParameters(
            paradigmParameters=GateModelParameters(qubitCount=circuit.qubit_count),
            errorMitigation=[error_mitigation.Debias()],
        ),
    )


@pytest.mark.parametrize("device_arn,device_parameters_class", DEVICE_PARAMETERS)
def test_create_circuit_with_input_params(
    device_arn, device_parameters_class, aws_session, circuit
):
    mocked_task_arn = "task-arn-1"
    aws_session.create_quantum_task.return_value = mocked_task_arn
    shots = 53
    inputs = {"beta": 3}

    task = AwsQuantumTask.create(aws_session, device_arn, circuit, S3_TARGET, shots, inputs=inputs)
    assert task == AwsQuantumTask(mocked_task_arn, aws_session)
    program = circuit.to_ir(ir_type=IRType.OPENQASM)
    assert program.inputs == {}
    program.inputs.update(inputs)

    _assert_create_quantum_task_called_with(
        aws_session,
        device_arn,
        program.json(),
        S3_TARGET,
        shots,
        device_parameters_class(
            paradigmParameters=GateModelParameters(
                qubitCount=circuit.qubit_count, disableQubitRewiring=False
            )
        ),
    )


@pytest.mark.parametrize(
    "device_arn,device_parameters_class", [(RIGETTI_ARN, RigettiDeviceParameters)]
)
def test_create_circuit_with_disabled_rewiring(
    device_arn, device_parameters_class, aws_session, circuit
):
    mocked_task_arn = "task-arn-1"
    aws_session.create_quantum_task.return_value = mocked_task_arn
    shots = 53

    task = AwsQuantumTask.create(
        aws_session, device_arn, circuit, S3_TARGET, shots, disable_qubit_rewiring=True
    )
    assert task == AwsQuantumTask(mocked_task_arn, aws_session)

    _assert_create_quantum_task_called_with(
        aws_session,
        device_arn,
        circuit.to_ir(
            ir_type=IRType.OPENQASM,
            serialization_properties=OpenQASMSerializationProperties(QubitReferenceType.PHYSICAL),
        ).json(),
        S3_TARGET,
        shots,
        device_parameters_class(
            paradigmParameters=GateModelParameters(
                qubitCount=circuit.qubit_count, disableQubitRewiring=True
            )
        ),
    )


@pytest.mark.parametrize(
    "device_arn,device_parameters_class, disable_qubit_rewiring",
    [(RIGETTI_ARN, RigettiDeviceParameters, True), (RIGETTI_ARN, RigettiDeviceParameters, False)],
)
def test_create_circuit_with_verbatim(
    device_arn, device_parameters_class, disable_qubit_rewiring, aws_session
):
    circ = Circuit().add_verbatim_box(Circuit().h(0))
    mocked_task_arn = "task-arn-1"
    aws_session.create_quantum_task.return_value = mocked_task_arn
    shots = 1337

    task = AwsQuantumTask.create(
        aws_session,
        device_arn,
        circ,
        S3_TARGET,
        shots,
        disable_qubit_rewiring=disable_qubit_rewiring,
    )
    assert task == AwsQuantumTask(mocked_task_arn, aws_session)

    serialization_properties = OpenQASMSerializationProperties(
        qubit_reference_type=QubitReferenceType.PHYSICAL
    )

    _assert_create_quantum_task_called_with(
        aws_session,
        device_arn,
        circ.to_ir(
            ir_type=IRType.OPENQASM,
            serialization_properties=serialization_properties,
        ).json(),
        S3_TARGET,
        shots,
        device_parameters_class(
            paradigmParameters=GateModelParameters(
                qubitCount=circ.qubit_count, disableQubitRewiring=disable_qubit_rewiring
            )
        ),
    )


@pytest.mark.xfail(raises=ValueError)
def test_create_circuit_with_shots_value_error(aws_session, arn, circuit):
    mocked_task_arn = "task-arn-1"
    aws_session.create_quantum_task.return_value = mocked_task_arn
    AwsQuantumTask.create(aws_session, arn, circuit, S3_TARGET, 0)


@pytest.mark.parametrize(
    "device_parameters,arn",
    [
        (
            {
                "providerLevelParameters": {
                    "postprocessingType": "OPTIMIZATION",
                    "annealingOffsets": [3.67, 6.123],
                    "annealingSchedule": [[13.37, 10.08], [3.14, 1.618]],
                    "annealingDuration": 1,
                    "autoScale": False,
                    "beta": 0.2,
                    "chains": [[0, 1, 5], [6]],
                    "compensateFluxDrift": False,
                    "fluxBiases": [1.1, 2.2, 3.3, 4.4],
                    "initialState": [1, 3, 0, 1],
                    "maxResults": 1,
                    "programmingThermalizationDuration": 625,
                    "readoutThermalizationDuration": 256,
                    "reduceIntersampleCorrelation": False,
                    "reinitializeState": True,
                    "resultFormat": "RAW",
                    "spinReversalTransformCount": 100,
                }
            },
            "arn:aws:braket:::device/qpu/d-wave/Advantage_system1",
        ),
        (
            {
                "deviceLevelParameters": {
                    "postprocessingType": "OPTIMIZATION",
                    "beta": 0.2,
                    "annealingOffsets": [3.67, 6.123],
                    "annealingSchedule": [[13.37, 10.08], [3.14, 1.618]],
                    "annealingDuration": 1,
                    "autoScale": False,
                    "chains": [[0, 1, 5], [6]],
                    "compensateFluxDrift": False,
                    "fluxBiases": [1.1, 2.2, 3.3, 4.4],
                    "initialState": [1, 3, 0, 1],
                    "maxResults": 1,
                    "programmingThermalizationDuration": 625,
                    "readoutThermalizationDuration": 256,
                    "reduceIntersampleCorrelation": False,
                    "reinitializeState": True,
                    "resultFormat": "RAW",
                    "spinReversalTransformCount": 100,
                }
            },
            "arn:aws:braket:::device/qpu/d-wave/DW_2000Q_6",
        ),
        pytest.param(
            {
                "deviceLevelParameters": {
                    "postprocessingType": "OPTIMIZATION",
                    "beta": 0.2,
                    "annealingOffsets": [3.67, 6.123],
                    "annealingSchedule": [[13.37, 10.08], [3.14, 1.618]],
                    "annealingDuration": 1,
                    "autoScale": False,
                    "chains": [[0, 1, 5], [6]],
                    "compensateFluxDrift": False,
                    "fluxBiases": [1.1, 2.2, 3.3, 4.4],
                    "initialState": [1, 3, 0, 1],
                    "maxResults": 1,
                    "programmingThermalizationDuration": 625,
                    "readoutThermalizationDuration": 256,
                    "reduceIntersampleCorrelation": False,
                    "reinitializeState": True,
                    "resultFormat": "RAW",
                    "spinReversalTransformCount": 100,
                }
            },
            "arn:aws:braket:::device/qpu/d-wave/Advantage_system1",
            # this doesn't fail... yet
            # marks=pytest.mark.xfail(reason='beta not a valid parameter for Advantage device'),
        ),
        pytest.param(
            {
                "deviceLevelParameters": {
                    "postprocessingType": "OPTIMIZATION",
                    "beta": 0.2,
                    "annealingOffsets": [3.67, 6.123],
                    "annealingSchedule": [[13.37, 10.08], [3.14, 1.618]],
                    "annealingDuration": 1,
                    "autoScale": False,
                    "chains": [[0, 1, 5], [6]],
                    "compensateFluxDrift": False,
                    "fluxBiases": [1.1, 2.2, 3.3, 4.4],
                    "initialState": [1, 3, 0, 1],
                    "maxResults": 1,
                    "programmingThermalizationDuration": 625,
                    "readoutThermalizationDuration": 256,
                    "reduceIntersampleCorrelation": False,
                    "reinitializeState": True,
                    "resultFormat": "RAW",
                    "spinReversalTransformCount": 100,
                }
            },
            "arn:aws:braket:::device/qpu/d-wave/fake_arn",
            marks=pytest.mark.xfail(reason="Bad ARN"),
        ),
        (
            {
                "deviceLevelParameters": {
                    "postprocessingType": "OPTIMIZATION",
                    "annealingOffsets": [3.67, 6.123],
                    "annealingSchedule": [[13.37, 10.08], [3.14, 1.618]],
                    "annealingDuration": 1,
                    "autoScale": False,
                    "beta": 0.2,
                    "chains": [[0, 1, 5], [6]],
                    "compensateFluxDrift": False,
                    "fluxBiases": [1.1, 2.2, 3.3, 4.4],
                    "initialState": [1, 3, 0, 1],
                    "maxResults": 1,
                    "programmingThermalizationDuration": 625,
                    "readoutThermalizationDuration": 256,
                    "reduceIntersampleCorrelation": False,
                    "reinitializeState": True,
                    "resultFormat": "RAW",
                    "spinReversalTransformCount": 100,
                }
            },
            "arn:aws:braket:::device/qpu/d-wave/DW_2000Q_6",
        ),
        (
            DwaveDeviceParameters.parse_obj(
                {
                    "providerLevelParameters": {
                        "postprocessingType": "OPTIMIZATION",
                        "annealingOffsets": [3.67, 6.123],
                        "annealingSchedule": [[13.37, 10.08], [3.14, 1.618]],
                        "annealingDuration": 1,
                        "autoScale": False,
                        "beta": 0.2,
                        "chains": [[0, 1, 5], [6]],
                        "compensateFluxDrift": False,
                        "fluxBiases": [1.1, 2.2, 3.3, 4.4],
                        "initialState": [1, 3, 0, 1],
                        "maxResults": 1,
                        "programmingThermalizationDuration": 625,
                        "readoutThermalizationDuration": 256,
                        "reduceIntersampleCorrelation": False,
                        "reinitializeState": True,
                        "resultFormat": "RAW",
                        "spinReversalTransformCount": 100,
                    }
                }
            ),
            "arn:aws:braket:::device/qpu/d-wave/Advantage_system1",
        ),
        (
            DwaveDeviceParameters.parse_obj(
                {
                    "deviceLevelParameters": {
                        "postprocessingType": "OPTIMIZATION",
                        "annealingOffsets": [3.67, 6.123],
                        "annealingSchedule": [[13.37, 10.08], [3.14, 1.618]],
                        "annealingDuration": 1,
                        "autoScale": False,
                        "beta": 0.2,
                        "chains": [[0, 1, 5], [6]],
                        "compensateFluxDrift": False,
                        "fluxBiases": [1.1, 2.2, 3.3, 4.4],
                        "initialState": [1, 3, 0, 1],
                        "maxResults": 1,
                        "programmingThermalizationDuration": 625,
                        "readoutThermalizationDuration": 256,
                        "reduceIntersampleCorrelation": False,
                        "reinitializeState": True,
                        "resultFormat": "RAW",
                        "spinReversalTransformCount": 100,
                    }
                },
            ),
            "arn:aws:braket:::device/qpu/d-wave/Advantage_system1",
        ),
        (
            DwaveAdvantageDeviceParameters.parse_obj(
                {
                    "deviceLevelParameters": {
                        "annealingOffsets": [3.67, 6.123],
                        "annealingSchedule": [[13.37, 10.08], [3.14, 1.618]],
                        "annealingDuration": 1,
                        "autoScale": False,
                        "beta": 0.2,
                        "chains": [[0, 1, 5], [6]],
                        "compensateFluxDrift": False,
                        "fluxBiases": [1.1, 2.2, 3.3, 4.4],
                        "initialState": [1, 3, 0, 1],
                        "maxResults": 1,
                        "programmingThermalizationDuration": 625,
                        "readoutThermalizationDuration": 256,
                        "reduceIntersampleCorrelation": False,
                        "reinitializeState": True,
                        "resultFormat": "RAW",
                        "spinReversalTransformCount": 100,
                    }
                },
            ),
            "arn:aws:braket:::device/qpu/d-wave/Advantage_system1",
        ),
        (
            Dwave2000QDeviceParameters.parse_obj(
                {
                    "deviceLevelParameters": {
                        "postprocessingType": "OPTIMIZATION",
                        "annealingOffsets": [3.67, 6.123],
                        "annealingSchedule": [[13.37, 10.08], [3.14, 1.618]],
                        "annealingDuration": 1,
                        "autoScale": False,
                        "beta": 0.2,
                        "chains": [[0, 1, 5], [6]],
                        "compensateFluxDrift": False,
                        "fluxBiases": [1.1, 2.2, 3.3, 4.4],
                        "initialState": [1, 3, 0, 1],
                        "maxResults": 1,
                        "programmingThermalizationDuration": 625,
                        "readoutThermalizationDuration": 256,
                        "reduceIntersampleCorrelation": False,
                        "reinitializeState": True,
                        "resultFormat": "RAW",
                        "spinReversalTransformCount": 100,
                    }
                }
            ),
            "arn:aws:braket:::device/qpu/d-wave/DW_2000Q_6",
        ),
        (
            Dwave2000QDeviceParameters.parse_obj({"deviceLevelParameters": {}}),
            "arn:aws:braket:::device/qpu/d-wave/DW_2000Q_6",
        ),
        pytest.param(
            {},
            "arn:aws:braket:::device/qpu/d-wave/DW_2000Q_6",
        ),
    ],
)
def test_from_annealing(device_parameters, aws_session, arn, problem):
    mocked_task_arn = "task-arn-1"
    aws_session.create_quantum_task.return_value = mocked_task_arn
    task = AwsQuantumTask.create(
        aws_session,
        arn,
        problem,
        S3_TARGET,
        1000,
        device_parameters=device_parameters,
    )
    assert task == AwsQuantumTask(mocked_task_arn, aws_session)
    annealing_parameters = _create_annealing_device_params(device_parameters, device_arn=arn)
    validate(
        json.loads(annealing_parameters.json(exclude_none=True)), annealing_parameters.schema()
    )
    _assert_create_quantum_task_called_with(
        aws_session,
        arn,
        problem.to_ir().json(),
        S3_TARGET,
        1000,
        annealing_parameters,
    )


@pytest.mark.parametrize("device_arn,device_parameters_class", DEVICE_PARAMETERS)
def test_create_with_tags(device_arn, device_parameters_class, aws_session, circuit):
    mocked_task_arn = "task-arn-tags"
    aws_session.create_quantum_task.return_value = mocked_task_arn
    shots = 53
    tags = {"state": "washington"}

    task = AwsQuantumTask.create(aws_session, device_arn, circuit, S3_TARGET, shots, tags=tags)
    assert task == AwsQuantumTask(mocked_task_arn, aws_session)
    _assert_create_quantum_task_called_with(
        aws_session,
        device_arn,
        circuit.to_ir(ir_type=IRType.OPENQASM).json(),
        S3_TARGET,
        shots,
        device_parameters_class(
            paradigmParameters=GateModelParameters(qubitCount=circuit.qubit_count)
        ),
        tags,
    )


def test_init_new_thread(aws_session, arn):
    tasks_list = []
    threading.Thread(target=_init_and_add_to_list, args=(aws_session, arn, tasks_list)).start()
    time.sleep(0.1)
    assert len(tasks_list) == 1


@patch("braket.aws.aws_quantum_task.boto3.Session")
def test_aws_session_for_task_arn(mock_session):
    region = "us-west-2"
    arn = f"arn:aws:aqx:{region}:account_id:quantum-task:task_id"
    mock_boto_session = Mock()
    mock_session.return_value = mock_boto_session
    mock_boto_session.region_name = region
    aws_session = AwsQuantumTask._aws_session_for_task_arn(arn)
    mock_session.assert_called_with(region_name=region)
    assert aws_session.boto_session == mock_boto_session


def _init_and_add_to_list(aws_session, arn, task_list):
    task_list.append(AwsQuantumTask(arn, aws_session))


def _assert_create_quantum_task_called_with(
    aws_session,
    arn,
    task_description,
    s3_results_prefix,
    shots,
    device_parameters=None,
    tags=None,
    reservation_arn=None,
):
    test_kwargs = {
        "deviceArn": arn,
        "outputS3Bucket": s3_results_prefix[0],
        "outputS3KeyPrefix": s3_results_prefix[1],
        "action": task_description,
        "shots": shots,
    }

    if device_parameters is not None:
        test_kwargs.update({"deviceParameters": device_parameters.json(exclude_none=True)})
    if tags is not None:
        test_kwargs.update({"tags": tags})
    if reservation_arn:
        test_kwargs.update(
            {
                "associations": [
                    {
                        "arn": reservation_arn,
                        "type": "RESERVATION_TIME_WINDOW_ARN",
                    }
                ]
            }
        )
    aws_session.create_quantum_task.assert_called_with(**test_kwargs)


def _mock_metadata(aws_session, state):
    message = (
        f"'Task is in {state} status. AmazonBraket does not show queue position for this status.'"
    )
    if state in AwsQuantumTask.TERMINAL_STATES or state in ["RUNNING", "CANCELLING"]:
        aws_session.get_quantum_task.return_value = {
            "status": state,
            "outputS3Bucket": S3_TARGET.bucket,
            "outputS3Directory": S3_TARGET.key,
            "queueInfo": {
                "queue": "QUANTUM_TASKS_QUEUE",
                "position": "None",
                "queuePriority": "Normal",
                "message": message,
            },
        }
    else:
        aws_session.get_quantum_task.return_value = {
            "status": state,
            "outputS3Bucket": S3_TARGET.bucket,
            "outputS3Directory": S3_TARGET.key,
            "queueInfo": {
                "queue": "QUANTUM_TASKS_QUEUE",
                "position": "2",
                "queuePriority": "Normal",
            },
        }


def _mock_s3(aws_session, result):
    aws_session.retrieve_s3_object_body.return_value = result


@pytest.mark.parametrize("source_input", [{}, {"gamma": 0.15}, None])
@pytest.mark.parametrize("device_arn", DEVICE_PARAMETERS[0])
def test_program_inputs(aws_session, device_arn, source_input):
    bell_qasm = """
    OPENQASM 3;
    input float theta;
    qubit[2] q;
    bit[2] c;
    h q[0];
    cnot q[0], q[1];
    c = measure q;
    """
    openqasm_program = OpenQASMProgram(source=bell_qasm, inputs=source_input)
    aws_session.create_quantum_task.return_value = arn
    shots = 21
    inputs = {"theta": 0.2}
    AwsQuantumTask.create(
        aws_session, device_arn, openqasm_program, S3_TARGET, shots, inputs=inputs
    )

    assert openqasm_program.inputs == source_input
    inputs_copy = openqasm_program.inputs.copy() if openqasm_program.inputs is not None else {}
    inputs_copy.update(inputs)
    openqasm_program = OpenQASMProgram(source=openqasm_program.source, inputs=inputs_copy)

    _assert_create_quantum_task_called_with(
        aws_session,
        device_arn,
        openqasm_program.json(),
        S3_TARGET,
        shots,
    )



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import json
import tempfile
from collections.abc import Iterable
from logging import Logger, getLogger
from pathlib import Path
from typing import Any

from braket.aws.aws_session import AwsSession
from braket.jobs.local.local_job_container import _LocalJobContainer


def setup_container(
    container: _LocalJobContainer, aws_session: AwsSession, **creation_kwargs: str
) -> dict[str, str]:
    """Sets up a container with prerequisites for running a Braket Hybrid Job. The prerequisites are
    based on the options the customer has chosen for the hybrid job. Similarly, any environment
    variables that are needed during runtime will be returned by this function.

    Args:
        container(_LocalJobContainer): The container that will run the braket hybrid job.
        aws_session (AwsSession): AwsSession for connecting to AWS Services.
        **creation_kwargs (str): Arbitrary keyword arguments.

    Returns:
        dict[str, str]: A dictionary of environment variables that reflect Braket Hybrid Jobs
        options requested by the customer.
    """
    logger = getLogger(__name__)
    _create_expected_paths(container, **creation_kwargs)
    run_environment_variables = {}
    run_environment_variables.update(_get_env_credentials(aws_session, logger))
    run_environment_variables.update(
        _get_env_script_mode_config(creation_kwargs["algorithmSpecification"]["scriptModeConfig"])
    )
    run_environment_variables.update(_get_env_default_vars(aws_session, **creation_kwargs))
    if _copy_hyperparameters(container, **creation_kwargs):
        run_environment_variables.update(_get_env_hyperparameters())
    if _copy_input_data_list(container, aws_session, **creation_kwargs):
        run_environment_variables.update(_get_env_input_data())
    return run_environment_variables


def _create_expected_paths(container: _LocalJobContainer, **creation_kwargs: str) -> None:
    """Creates the basic paths required for Braket Hybrid Jobs to run.

    Args:
        container(_LocalJobContainer): The container that will run the braket hybrid job.
        **creation_kwargs (str): Arbitrary keyword arguments.
    """
    container.makedir("/opt/ml/model")
    container.makedir(creation_kwargs["checkpointConfig"]["localPath"])


def _get_env_credentials(aws_session: AwsSession, logger: Logger) -> dict[str, str]:
    """Gets the account credentials from boto so they can be added as environment variables to
    the running container.

    Args:
        aws_session (AwsSession): AwsSession for connecting to AWS Services.
        logger (Logger): Logger object with which to write logs. Default is `getLogger(__name__)`

    Returns:
        dict[str, str]: The set of key/value pairs that should be added as environment variables
        to the running container.
    """
    credentials = aws_session.boto_session.get_credentials()
    if credentials.token is None:
        logger.info("Using the long-lived AWS credentials found in session")
        return {
            "AWS_ACCESS_KEY_ID": str(credentials.access_key),
            "AWS_SECRET_ACCESS_KEY": str(credentials.secret_key),
        }
    logger.warning(
        "Using the short-lived AWS credentials found in session. They might expire while running."
    )
    return {
        "AWS_ACCESS_KEY_ID": str(credentials.access_key),
        "AWS_SECRET_ACCESS_KEY": str(credentials.secret_key),
        "AWS_SESSION_TOKEN": str(credentials.token),
    }


def _get_env_script_mode_config(script_mode_config: dict[str, str]) -> dict[str, str]:
    """Gets the environment variables related to the customer script mode config.

    Args:
        script_mode_config (dict[str, str]): The values for scriptModeConfig in the boto3 input
            parameters for running a Braket Hybrid Job.

    Returns:
        dict[str, str]: The set of key/value pairs that should be added as environment variables
        to the running container.
    """
    result = {
        "AMZN_BRAKET_SCRIPT_S3_URI": script_mode_config["s3Uri"],
        "AMZN_BRAKET_SCRIPT_ENTRY_POINT": script_mode_config["entryPoint"],
    }
    if "compressionType" in script_mode_config:
        result["AMZN_BRAKET_SCRIPT_COMPRESSION_TYPE"] = script_mode_config["compressionType"]
    return result


def _get_env_default_vars(aws_session: AwsSession, **creation_kwargs: str) -> dict[str, str]:
    """This function gets the remaining 'simple' env variables, that don't require any
     additional logic to determine what they are or when they should be added as env variables.

    Args:
        aws_session (AwsSession): AwsSession for connecting to AWS Services.
        **creation_kwargs (str): Arbitrary keyword arguments.

    Returns:
        dict[str, str]: The set of key/value pairs that should be added as environment variables
        to the running container.
    """
    job_name = creation_kwargs["jobName"]
    bucket, location = AwsSession.parse_s3_uri(creation_kwargs["outputDataConfig"]["s3Path"])
    return {
        "AWS_DEFAULT_REGION": aws_session.region,
        "AMZN_BRAKET_JOB_NAME": job_name,
        "AMZN_BRAKET_DEVICE_ARN": creation_kwargs["deviceConfig"]["device"],
        "AMZN_BRAKET_JOB_RESULTS_DIR": "/opt/braket/model",
        "AMZN_BRAKET_CHECKPOINT_DIR": creation_kwargs["checkpointConfig"]["localPath"],
        "AMZN_BRAKET_OUT_S3_BUCKET": bucket,
        "AMZN_BRAKET_TASK_RESULTS_S3_URI": f"s3://{bucket}/jobs/{job_name}/tasks",
        "AMZN_BRAKET_JOB_RESULTS_S3_PATH": str(Path(location, job_name, "output").as_posix()),
    }


def _get_env_hyperparameters() -> dict[str, str]:
    """Gets the env variable for hyperparameters. This should only be added if the customer has
    provided hyperpameters to the hybrid job.

    Returns:
        dict[str, str]: The set of key/value pairs that should be added as environment variables
        to the running container.
    """
    return {
        "AMZN_BRAKET_HP_FILE": "/opt/braket/input/config/hyperparameters.json",
    }


def _get_env_input_data() -> dict[str, str]:
    """Gets the env variable for input data. This should only be added if the customer has
    provided input data to the hybrid job.

    Returns:
        dict[str, str]: The set of key/value pairs that should be added as environment variables
        to the running container.
    """
    return {
        "AMZN_BRAKET_INPUT_DIR": "/opt/braket/input/data",
    }


def _copy_hyperparameters(container: _LocalJobContainer, **creation_kwargs: str) -> bool:
    """If hyperpameters are present, this function will store them as a JSON object in the
     container in the appropriate location on disk.

    Args:
        container(_LocalJobContainer): The container to save hyperparameters to.
        **creation_kwargs (str): Arbitrary keyword arguments.

    Returns:
        bool: True if any hyperparameters were copied to the container.
    """
    if "hyperParameters" not in creation_kwargs:
        return False
    hyperparameters = creation_kwargs["hyperParameters"]
    with tempfile.TemporaryDirectory() as temp_dir:
        file_path = Path(temp_dir, "hyperparameters.json")
        with open(file_path, "w") as write_file:
            json.dump(hyperparameters, write_file)
        container.copy_to(str(file_path), "/opt/ml/input/config/hyperparameters.json")
    return True


def _download_input_data(
    aws_session: AwsSession,
    download_dir: str,
    input_data: dict[str, Any],
) -> None:
    """Downloads input data for a hybrid job.

    Args:
        aws_session (AwsSession): AwsSession for connecting to AWS Services.
        download_dir (str): The directory path to download to.
        input_data (dict[str, Any]): One of the input data in the boto3 input parameters for
            running a Braket Hybrid Job.

    Raises:
        ValueError: File already exists.
        RuntimeError: The item is not found.

    """
    # If s3 prefix is the full name of a directory and all keys are inside
    # that directory, the contents of said directory will be copied into a
    # directory with the same name as the channel. This behavior is the same
    # whether or not s3 prefix ends with a "/". Moreover, if s3 prefix ends
    # with a "/", this is certainly the behavior to expect, since it can only
    # match a directory.
    # If s3 prefix matches any files exactly, or matches as a prefix of any
    # files or directories, then all files and directories matching s3 prefix
    # will be copied into a directory with the same name as the channel.
    channel_name = input_data["channelName"]
    s3_uri_prefix = input_data["dataSource"]["s3DataSource"]["s3Uri"]
    bucket, prefix = AwsSession.parse_s3_uri(s3_uri_prefix)
    s3_keys = aws_session.list_keys(bucket, prefix)
    top_level = prefix if _is_dir(prefix, s3_keys) else str(Path(prefix).parent)
    found_item = False
    try:
        Path(download_dir, channel_name).mkdir()
    except FileExistsError:
        raise ValueError(f"Duplicate channel names not allowed for input data: {channel_name}")
    for s3_key in s3_keys:
        relative_key = Path(s3_key).relative_to(top_level)
        download_path = Path(download_dir, channel_name, relative_key)
        if not s3_key.endswith("/"):
            download_path.parent.mkdir(parents=True, exist_ok=True)
            aws_session.download_from_s3(
                AwsSession.construct_s3_uri(bucket, s3_key), str(download_path)
            )
            found_item = True
    if not found_item:
        raise RuntimeError(f"No data found for channel '{channel_name}'")


def _is_dir(prefix: str, keys: Iterable[str]) -> bool:
    """Determine whether the prefix refers to a directory.

    Args:
        prefix (str): The prefix to check.
        keys (Iterable[str]): The set of paths to check.

    Returns:
        bool: True if the prefix refers to a directory.
    """
    if prefix.endswith("/"):
        return True
    return all(key.startswith(f"{prefix}/") for key in keys)


def _copy_input_data_list(
    container: _LocalJobContainer, aws_session: AwsSession, **creation_kwargs: str
) -> bool:
    """If the input data list is not empty, this function will download the input files and
    store them in the container.

    Args:
        container (_LocalJobContainer): The container to save input data to.
        aws_session (AwsSession): AwsSession for connecting to AWS Services.
        **creation_kwargs (str): Arbitrary keyword arguments.

    Returns:
        bool: True if any input data was copied to the container.
    """
    if "inputDataConfig" not in creation_kwargs:
        return False

    input_data_list = creation_kwargs["inputDataConfig"]
    with tempfile.TemporaryDirectory() as temp_dir:
        for input_data in input_data_list:
            _download_input_data(aws_session, temp_dir, input_data)
        container.copy_to(temp_dir, "/opt/ml/input/data/")
    return bool(input_data_list)



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from typing import List, Set

import pytest

from braket.aws import AwsDevice
from braket.devices import Devices

RIGETTI_ARN = "arn:aws:braket:us-west-1::device/qpu/rigetti/Aspen-M-3"
IONQ_ARN = "arn:aws:braket:us-east-1::device/qpu/ionq/Harmony"
SIMULATOR_ARN = "arn:aws:braket:::device/quantum-simulator/amazon/sv1"
OQC_ARN = "arn:aws:braket:eu-west-2::device/qpu/oqc/Lucy"
PULSE_ARN = "arn:aws:braket:us-west-1::device/qpu/rigetti/Aspen-M-3"


@pytest.mark.parametrize(
    "arn", [(RIGETTI_ARN), (IONQ_ARN), (OQC_ARN), (SIMULATOR_ARN), (PULSE_ARN)]
)
def test_device_creation(arn, created_braket_devices):
    device = created_braket_devices[arn]
    assert device.arn == arn
    assert device.name
    assert device.status
    assert device.type
    assert device.provider_name
    assert device.properties


@pytest.mark.parametrize("arn", [(PULSE_ARN)])
def test_device_pulse_properties(arn, aws_session, created_braket_devices):
    device = created_braket_devices[arn]
    assert device.ports
    assert device.frames


def test_device_across_regions(aws_session, created_braket_devices):
    # assert QPUs across different regions can be created using the same aws_session
    created_braket_devices[RIGETTI_ARN]
    created_braket_devices[IONQ_ARN]
    created_braket_devices[OQC_ARN]


@pytest.mark.parametrize("arn", [(RIGETTI_ARN), (IONQ_ARN), (OQC_ARN), (SIMULATOR_ARN)])
def test_get_devices_arn(arn):
    results = AwsDevice.get_devices(arns=[arn])
    assert results[0].arn == arn


@pytest.mark.parametrize("arn", [(PULSE_ARN)])
def test_device_gate_calibrations(arn, aws_session, created_braket_devices):
    device = created_braket_devices[arn]
    assert device.gate_calibrations


def test_get_devices_others():
    provider_names = ["Amazon Braket"]
    types = ["SIMULATOR"]
    statuses = ["ONLINE"]
    results = AwsDevice.get_devices(provider_names=provider_names, types=types, statuses=statuses)
    assert results
    for result in results:
        assert result.provider_name in provider_names
        assert result.type in types
        assert result.status in statuses


def test_get_devices_all(braket_devices):
    result_arns = [result.arn for result in braket_devices]
    for arn in [RIGETTI_ARN, IONQ_ARN, SIMULATOR_ARN, OQC_ARN]:
        assert arn in result_arns


def _get_provider_name(device: AwsDevice) -> str:
    arn_provider = device.arn.split("/")[2]

    # capitalize as in provider name
    provider_primary_name = device.provider_name.split()[0]
    if arn_provider == provider_primary_name.lower():
        capitalized = provider_primary_name
    else:
        capitalized = arn_provider.upper()

    # remove dashes
    return capitalized.replace("-", "")


def _get_device_name(device: AwsDevice) -> str:
    arn_device_name = device.arn.split("/")[-1]

    device_name = (
        arn_device_name.replace("Advantage_system", "Advantage").replace("_", "").replace("-", "")
    )

    if device.provider_name == "Amazon Braket":
        device_name = device_name.upper()
    return device_name


def _get_active_providers(aws_devices: List[AwsDevice]) -> Set[str]:
    active_providers = set()
    for device in aws_devices:
        if device.status != "RETIRED":
            active_providers.add(_get_provider_name(device))
    return active_providers


def _validate_device(device: AwsDevice, active_providers: Set[str]):
    provider_name = _get_provider_name(device)
    if provider_name not in active_providers:
        provider_name = f"_{provider_name}"
    device_name = _get_device_name(device)
    if device.status == "RETIRED":
        device_name = f"_{device_name}"

    assert getattr(getattr(Devices, provider_name), device_name) == device.arn


def test_device_enum(braket_devices, created_braket_devices):
    active_providers = _get_active_providers(braket_devices)

    # validate all devices in API
    for device in braket_devices:
        _validate_device(device, active_providers)

    # validate all devices in enum
    providers = [getattr(Devices, attr) for attr in dir(Devices) if not attr.startswith("__")]
    for provider in providers:
        for device_arn in provider:
            device = created_braket_devices[device_arn]
            _validate_device(device, active_providers)



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from __future__ import annotations

from collections.abc import Iterable
from typing import Any, Union

from boltons.setutils import IndexedSet

from braket.registers.qubit import Qubit, QubitInput

QubitSetInput = Union[QubitInput, Iterable[QubitInput]]


def _flatten(other: Any) -> Any:
    if isinstance(other, Iterable) and not isinstance(other, str):
        for item in other:
            yield from _flatten(item)
    else:
        yield other


class QubitSet(IndexedSet):
    """An ordered, unique set of quantum bits.

    Note:
        QubitSet implements `__hash__()` but is a mutable object, therefore be careful when
        mutating this object.
    """

    def __init__(self, qubits: QubitSetInput | None = None):
        """Initializes a `QubitSet`.

        Args:
            qubits (QubitSetInput | None): Qubits to be included in the `QubitSet`.
                Default is `None`.

        Examples:
            >>> qubits = QubitSet([0, 1])
            >>> for qubit in qubits:
            ...     print(qubit)
            ...
            Qubit(0)
            Qubit(1)

            >>> qubits = QubitSet([0, 1, [2, 3]])
            >>> for qubit in qubits:
            ...     print(qubit)
            ...
            Qubit(0)
            Qubit(1)
            Qubit(2)
            Qubit(3)
        """
        _qubits = [Qubit.new(qubit) for qubit in _flatten(qubits)] if qubits is not None else None
        super().__init__(_qubits)

    def map(self, mapping: dict[QubitInput, QubitInput]) -> QubitSet:
        """Creates a new `QubitSet` where this instance's qubits are mapped to the values in `mapping`.
        If this instance contains a qubit that is not in the `mapping` that qubit is not modified.

        Args:
            mapping (dict[QubitInput, QubitInput]): A dictionary of qubit mappings to
                apply. Key is the qubit in this instance to target, and the value is what
                the key will be changed to.

        Returns:
            QubitSet: A new QubitSet with the `mapping` applied.

        Examples:
            >>> qubits = QubitSet([0, 1])
            >>> mapping = {0: 10, Qubit(1): Qubit(11)}
            >>> qubits.map(mapping)
            QubitSet([Qubit(10), Qubit(11)])
        """  # noqa E501
        new_qubits = [mapping.get(qubit, qubit) for qubit in self]

        return QubitSet(new_qubits)

    def __hash__(self):
        return hash(tuple(self))



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from braket.aws import AwsDevice
from braket.circuits import Circuit
from braket.jobs import (
    get_hyperparameters,
    get_job_device_arn,
    save_job_checkpoint,
    save_job_result,
)
from braket.jobs_data import PersistedJobDataFormat


def start_here():
    hyperparameters = get_hyperparameters()

    if hyperparameters["test_case"] == "completed":
        completed_job_script()
    else:
        failed_job_script()


def failed_job_script():
    print("Test job started!!!!!")
    open("fake_file")


def completed_job_script():
    print("Test job started!!!!!")

    # Use the device declared in the Orchestration Script
    device = AwsDevice(get_job_device_arn())

    bell = Circuit().h(0).cnot(0, 1)
    for count in range(3):
        task = device.run(bell, shots=10)
        print(task.result().measurement_counts)
        save_job_result({"converged": True, "energy": -0.2})
        save_job_checkpoint({"some_data": "abc"}, checkpoint_file_suffix="plain_data")
        save_job_checkpoint({"some_data": "abc"}, data_format=PersistedJobDataFormat.PICKLED_V4)

    print("Test job completed!!!!!")


def job_helper():
    print("import successful!")
    return {"status": "SUCCESS"}



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import pytest

from braket.circuits import Gate, QubitSet
from braket.circuits.gate_calibrations import GateCalibrations
from braket.pulse import Frame, Port, PulseSequence


@pytest.fixture
def port():
    return Port("test_port_ff", dt=1e-9)


@pytest.fixture
def frame_id():
    return "test_frame_rf"


@pytest.fixture
def frame(frame_id, port):
    return Frame(frame_id, port, 1e6, is_predefined=True)


@pytest.fixture
def pulse_sequence(frame):
    return (
        PulseSequence()
        .barrier(qubits_or_frames=[frame])
        .delay(qubits_or_frames=[frame], duration=1000)
    )


def test_gc_creation(pulse_sequence):
    calibration_key = (Gate.H(), QubitSet([0, 1]))
    calibration = GateCalibrations({calibration_key: pulse_sequence})

    assert calibration.pulse_sequences[calibration_key] == pulse_sequence


def test_gc_copy(pulse_sequence):
    calibration_key = (Gate.H(), QubitSet([0, 1]))
    calibration = GateCalibrations({calibration_key: pulse_sequence})

    assert calibration == calibration.copy()


def test_filter(pulse_sequence):
    calibration_key = (Gate.Z(), QubitSet([0]))
    calibration_key_2 = (Gate.H(), QubitSet([1]))
    calibration_key_3 = (Gate.CZ(), QubitSet([0, 1]))
    calibration = GateCalibrations(
        {
            calibration_key: pulse_sequence,
            calibration_key_2: pulse_sequence,
            calibration_key_3: pulse_sequence,
        }
    )
    expected_calibration_1 = GateCalibrations({calibration_key: pulse_sequence})
    expected_calibration_2 = GateCalibrations(
        {calibration_key: pulse_sequence, calibration_key_3: pulse_sequence}
    )
    expected_calibration_3 = GateCalibrations({calibration_key_2: pulse_sequence})
    expected_calibration_4 = GateCalibrations({})
    expected_calibration_5 = calibration
    expected_calibration_6 = GateCalibrations({calibration_key_3: pulse_sequence})
    assert expected_calibration_1 == calibration.filter(gates=[Gate.Z()])
    assert expected_calibration_2 == calibration.filter(qubits=QubitSet(0))
    assert expected_calibration_3 == calibration.filter(gates=[Gate.H()], qubits=QubitSet(1))
    assert expected_calibration_4 == calibration.filter(gates=[Gate.Z()], qubits=QubitSet(1))
    assert expected_calibration_5 == calibration.filter(qubits=[QubitSet(0), QubitSet(1)])
    assert expected_calibration_6 == calibration.filter(qubits=QubitSet([0, 1]))


def test_to_ir(pulse_sequence):
    calibration_key = (Gate.Rx(angle=1), QubitSet([0, 1]))
    calibration = GateCalibrations({calibration_key: pulse_sequence})
    expected_ir = "\n".join(
        [
            "OPENQASM 3.0;",
            "defcal rx(1.0) $0, $1 {",
            "    barrier test_frame_rf;",
            "    delay[1000s] test_frame_rf;",
            "}",
        ]
    )

    assert calibration.to_ir() == expected_ir


@pytest.mark.xfail(raises=ValueError)
def test_to_ir_with_bad_key(pulse_sequence):
    calibration_key = (Gate.Z(), QubitSet([0, 1]))
    calibration_key_2 = (Gate.H(), QubitSet([0, 1]))
    calibration = GateCalibrations(
        {calibration_key: pulse_sequence, calibration_key_2: pulse_sequence}
    )
    expected_ir = "\n".join(
        [
            "OPENQASM 3.0;",
            "defcal z $0, $1 {",
            "    barrier test_frame_rf;",
            "    delay[1000s] test_frame_rf;",
            "}",
        ]
    )
    assert expected_ir == calibration.to_ir((Gate.Z(), QubitSet([1, 2])))


def test_to_ir_with_key(pulse_sequence):
    calibration_key = (Gate.Z(), QubitSet([0, 1]))
    calibration_key_2 = (Gate.H(), QubitSet([0, 1]))
    calibration = GateCalibrations(
        {calibration_key: pulse_sequence, calibration_key_2: pulse_sequence}
    )
    expected_ir = "\n".join(
        [
            "OPENQASM 3.0;",
            "defcal z $0, $1 {",
            "    barrier test_frame_rf;",
            "    delay[1000s] test_frame_rf;",
            "}",
        ]
    )
    assert expected_ir == calibration.to_ir(calibration_key)


def test_gate_calibrations_length(pulse_sequence):
    calibration_key = (Gate.Z(), QubitSet([0, 1]))
    calibration_key_2 = (Gate.H(), QubitSet([0, 1]))
    calibration = GateCalibrations(
        {calibration_key: pulse_sequence, calibration_key_2: pulse_sequence}
    )

    assert len(calibration) == 2


@pytest.mark.parametrize(
    "bad_input",
    [
        ({(Gate.Rx(1), "string"): PulseSequence()}),
        ({(Gate.Rx(1), QubitSet(0)): 4}),
        ({("string_a", "string_b"): PulseSequence()}),
    ],
)
@pytest.mark.xfail(raises=TypeError)
def test_bad_pulse_sequence(bad_input):
    GateCalibrations(bad_input)



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import numpy as np
import pytest

from braket.ir.annealing import Problem
from braket.task_result import (
    AdditionalMetadata,
    AnnealingTaskResult,
    DwaveMetadata,
    DwaveTiming,
    TaskMetadata,
)
from braket.tasks import AnnealingQuantumTaskResult


@pytest.fixture
def solutions():
    return [[-1, -1, -1, -1], [1, -1, 1, 1], [1, -1, -1, 1]]


@pytest.fixture
def values():
    return [0.0, 1.0, 2.0]


@pytest.fixture
def variable_count():
    return 4


@pytest.fixture
def solution_counts():
    return [3, 2, 4]


@pytest.fixture
def problem_type():
    return "ISING"


@pytest.fixture
def task_metadata():
    return TaskMetadata(**{"id": "task_arn", "deviceId": "arn1", "shots": 100})


@pytest.fixture
def dwave_metadata():
    return DwaveMetadata(
        activeVariables=[0],
        timing=DwaveTiming(
            qpuSamplingTime=100,
            qpuAnnealTimePerSample=20,
            qpuReadoutTimePerSample=274,
            qpuAccessTime=10917,
            qpuAccessOverheadTime=3382,
            qpuProgrammingTime=9342,
            qpuDelayTimePerSample=21,
            totalPostProcessingTime=117,
            postProcessingOverheadTime=117,
            totalRealTime=10917,
            runTimeChip=1575,
            annealTimePerRun=20,
            readoutTimePerRun=274,
        ),
    )


@pytest.fixture
def additional_metadata(problem_type, dwave_metadata):
    problem = Problem(
        type=problem_type,
        linear={0: 0.3333, 1: -0.333, 4: -0.333, 5: 0.333},
        quadratic={"0,4": 0.667, "0,5": -1, "1,4": 0.667, "1,5": 0.667},
    )
    return AdditionalMetadata(action=problem, dwaveMetadata=dwave_metadata)


@pytest.fixture
def result_str_1(
    solutions, values, solution_counts, variable_count, task_metadata, additional_metadata
):
    result = AnnealingTaskResult(
        solutions=solutions,
        variableCount=variable_count,
        values=values,
        solutionCounts=solution_counts,
        taskMetadata=task_metadata,
        additionalMetadata=additional_metadata,
    )
    return result.json()


@pytest.fixture
def result_str_2(solutions, values, variable_count, task_metadata, additional_metadata):
    result = AnnealingTaskResult(
        solutions=solutions,
        variableCount=variable_count,
        values=values,
        taskMetadata=task_metadata,
        additionalMetadata=additional_metadata,
    )
    return result.json()


@pytest.fixture
def result_str_3(solutions, values, variable_count, task_metadata, additional_metadata):
    result = AnnealingTaskResult(
        solutionCounts=[],
        solutions=solutions,
        variableCount=variable_count,
        values=values,
        taskMetadata=task_metadata,
        additionalMetadata=additional_metadata,
    )
    return result.json()


@pytest.fixture
def annealing_result(
    solutions,
    values,
    solution_counts,
    variable_count,
    problem_type,
    additional_metadata,
    task_metadata,
):
    solutions = np.asarray(solutions, dtype=int)
    values = np.asarray(values, dtype=float)
    solution_counts = np.asarray(solution_counts, dtype=int)
    record_array = AnnealingQuantumTaskResult._create_record_array(
        solutions, solution_counts, values
    )
    return AnnealingQuantumTaskResult(
        record_array=record_array,
        variable_count=variable_count,
        problem_type=problem_type,
        task_metadata=task_metadata,
        additional_metadata=additional_metadata,
    )


def test_from_object(
    result_str_1,
    solutions,
    values,
    solution_counts,
    variable_count,
    problem_type,
    task_metadata,
    additional_metadata,
):
    result = AnnealingQuantumTaskResult.from_object(AnnealingTaskResult.parse_raw(result_str_1))
    solutions = np.asarray(solutions, dtype=int)
    values = np.asarray(values, dtype=float)
    solution_counts = np.asarray(solution_counts, dtype=int)
    assert result.variable_count == variable_count
    assert result.problem_type == problem_type
    assert result.task_metadata == task_metadata
    assert result.additional_metadata == additional_metadata
    np.testing.assert_equal(
        result.record_array,
        AnnealingQuantumTaskResult._create_record_array(solutions, solution_counts, values),
    )


def test_from_string(
    result_str_1,
    solutions,
    values,
    solution_counts,
    variable_count,
    problem_type,
    task_metadata,
    additional_metadata,
):
    result = AnnealingQuantumTaskResult.from_string(result_str_1)
    solutions = np.asarray(solutions, dtype=int)
    values = np.asarray(values, dtype=float)
    solution_counts = np.asarray(solution_counts, dtype=int)
    assert result.variable_count == variable_count
    assert result.problem_type == problem_type
    assert result.task_metadata == task_metadata
    assert result.additional_metadata == additional_metadata
    np.testing.assert_equal(
        result.record_array,
        AnnealingQuantumTaskResult._create_record_array(solutions, solution_counts, values),
    )


def test_from_string_solution_counts_none(result_str_2, solutions):
    result = AnnealingQuantumTaskResult.from_string(result_str_2)
    np.testing.assert_equal(result.record_array.solution_count, np.ones(len(solutions), dtype=int))


def test_from_string_solution_counts_empty_list(result_str_3, solutions):
    result = AnnealingQuantumTaskResult.from_string(result_str_3)
    np.testing.assert_equal(result.record_array.solution_count, np.ones(len(solutions), dtype=int))


def test_data_sort_by_none(annealing_result, solutions, values, solution_counts):
    d = list(annealing_result.data(sorted_by=None))
    for i in range(len(solutions)):
        assert (d[i][0] == solutions[i]).all()
        assert d[i][1] == values[i]
        assert d[i][2] == solution_counts[i]


def test_data_selected_fields(annealing_result, solutions, values, solution_counts):
    d = list(annealing_result.data(selected_fields=["value"]))
    for i in range(len(solutions)):
        assert d[i] == tuple([values[i]])


def test_data_reverse(annealing_result, solutions, values, solution_counts):
    d = list(annealing_result.data(reverse=True))
    num_solutions = len(solutions)
    for i in range(num_solutions):
        assert (d[i][0] == solutions[num_solutions - i - 1]).all()
        assert d[i][1] == values[num_solutions - i - 1]
        assert d[i][2] == solution_counts[num_solutions - i - 1]


def test_data_sort_by(annealing_result, solutions, values, solution_counts):
    d = list(annealing_result.data(sorted_by="solution_count"))
    min_index = np.argmin(solution_counts)
    assert (d[0][0] == solutions[min_index]).all()
    assert d[0][1] == values[min_index]
    assert d[0][2] == solution_counts[min_index]


def test_from_object_equal_to_from_string(result_str_1):
    assert AnnealingQuantumTaskResult.from_object(
        AnnealingTaskResult.parse_raw(result_str_1)
    ) == AnnealingQuantumTaskResult.from_string(result_str_1)


def test_equality(result_str_1, result_str_2):
    result_1 = AnnealingQuantumTaskResult.from_string(result_str_1)
    result_2 = AnnealingQuantumTaskResult.from_string(result_str_1)
    other_result = AnnealingQuantumTaskResult.from_string(result_str_2)
    non_result = "not a quantum task result"

    assert result_1 == result_2
    assert result_1 is not result_2
    assert result_1 != other_result
    assert result_1 != non_result



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from __future__ import annotations

from braket.ahs.discretization_types import DiscretizationProperties
from braket.ahs.field import Field
from braket.ahs.hamiltonian import Hamiltonian
from braket.ahs.pattern import Pattern
from braket.timings.time_series import StitchBoundaryCondition, TimeSeries


class ShiftingField(Hamiltonian):
    def __init__(self, magnitude: Field) -> None:
        r"""Creates a Hamiltonian term :math:`H_{shift}` representing the shifting field
        that changes the energy of the Rydberg level in an AnalogHamiltonianSimulation,
        defined by the formula

        .. math::
            H_{shift} (t) := -\Delta(t) \sum_k h_k | r_k \rangle \langle r_k |

        where

            :math:`\Delta(t)` is the magnitude of the frequency shift in rad/s,

            :math:`h_k` is the site coefficient of atom :math:`k`,
            a dimensionless real number between 0 and 1,

            :math:`|r_k \rangle` is the Rydberg state of atom :math:`k`.

        with the sum :math:`\sum_k` taken over all target atoms.

        Args:
            magnitude (Field): containing the global magnitude time series :math:`\Delta(t)`,
                where time is measured in seconds (s) and values are measured in rad/s, and the
                local pattern :math:`h_k` of dimensionless real numbers between 0 and 1.
        """
        super().__init__()
        self._magnitude = magnitude

    @property
    def terms(self) -> list[Hamiltonian]:
        return [self]

    @property
    def magnitude(self) -> Field:
        r"""Field: containing the global magnitude time series :math:`\Delta(t)`,
        where time is measured in seconds (s) and values measured in rad/s)
        and the local pattern :math:`h_k` of dimensionless real numbers between 0 and 1.
        """
        return self._magnitude

    @staticmethod
    def from_lists(times: list[float], values: list[float], pattern: list[float]) -> ShiftingField:
        """Get the shifting field from a set of time points, values and pattern

        Args:
            times (list[float]): The time points of the shifting field
            values (list[float]): The values of the shifting field
            pattern (list[float]): The pattern of the shifting field

        Raises:
            ValueError: If the length of times and values differs.

        Returns:
            ShiftingField: The shifting field obtained
        """
        if len(times) != len(values):
            raise ValueError("The length of the times and values lists must be equal.")

        magnitude = TimeSeries()
        for t, v in zip(times, values):
            magnitude.put(t, v)
        shift = ShiftingField(Field(magnitude, Pattern(pattern)))

        return shift

    def stitch(
        self, other: ShiftingField, boundary: StitchBoundaryCondition = StitchBoundaryCondition.MEAN
    ) -> ShiftingField:
        """Stitches two shifting fields based on TimeSeries.stitch method.
        The time points of the second ShiftingField are shifted such that the first time point of
        the second ShiftingField coincides with the last time point of the first ShiftingField.
        The boundary point value is handled according to StitchBoundaryCondition argument value.

        Args:
            other (ShiftingField): The second shifting field to be stitched with.
            boundary (StitchBoundaryCondition): {"mean", "left", "right"}. Boundary point handler.

                Possible options are
              - "mean" - take the average of the boundary value points of the first
                and the second time series.
              - "left" - use the last value from the left time series as the boundary point.
              - "right" - use the first value from the right time series as the boundary point.

        Raises:
            ValueError: The ShiftingField patterns differ.

        Returns:
            ShiftingField: The stitched ShiftingField object.

        Example (StitchBoundaryCondition.MEAN):
        ::
            time_series_1 = TimeSeries.from_lists(times=[0, 0.1], values=[1, 2])
            time_series_2 = TimeSeries.from_lists(times=[0.2, 0.4], values=[4, 5])

            stitch_ts = time_series_1.stitch(time_series_2, boundary=StitchBoundaryCondition.MEAN)

            Result:
                stitch_ts.times() = [0, 0.1, 0.3]
                stitch_ts.values() = [1, 3, 5]

        Example (StitchBoundaryCondition.LEFT):
        ::
            stitch_ts = time_series_1.stitch(time_series_2, boundary=StitchBoundaryCondition.LEFT)

            Result:
                stitch_ts.times() = [0, 0.1, 0.3]
                stitch_ts.values() = [1, 2, 5]

        Example (StitchBoundaryCondition.RIGHT):
        ::
            stitch_ts = time_series_1.stitch(time_series_2, boundary=StitchBoundaryCondition.RIGHT)

            Result:
                stitch_ts.times() = [0, 0.1, 0.3]
                stitch_ts.values() = [1, 4, 5]
        """
        if not (self.magnitude.pattern.series == other.magnitude.pattern.series):
            raise ValueError("The ShiftingField pattern for both fields must be equal.")

        new_ts = self.magnitude.time_series.stitch(other.magnitude.time_series, boundary)
        return ShiftingField(Field(new_ts, self.magnitude.pattern))

    def discretize(self, properties: DiscretizationProperties) -> ShiftingField:
        """Creates a discretized version of the ShiftingField.

        Args:
            properties (DiscretizationProperties): Capabilities of a device that represent the
                resolution with which the device can implement the parameters.

        Returns:
            ShiftingField: A new discretized ShiftingField.
        """
        shifting_parameters = properties.rydberg.rydbergLocal
        discretized_magnitude = self.magnitude.discretize(
            time_resolution=shifting_parameters.timeResolution,
            value_resolution=shifting_parameters.commonDetuningResolution,
            pattern_resolution=shifting_parameters.localDetuningResolution,
        )
        return ShiftingField(discretized_magnitude)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from typing import Any, Optional

from openpulse import ast
from openqasm3.visitor import QASMTransformer


class _IRQASMTransformer(QASMTransformer):
    """QASMTransformer which walks the AST and makes the necessary modifications needed
    for IR generation. Currently, it performs the following operations:
      * Replaces capture_v0 function calls with assignment statements, assigning the
        readout value to a bit register element.
    """

    def __init__(self, register_identifier: Optional[str] = None):
        self._register_identifier = register_identifier
        self._capture_v0_count = 0
        super().__init__()

    def visit_ExpressionStatement(self, expression_statement: ast.ExpressionStatement) -> Any:
        """Visit an Expression.

        Args:
            expression_statement (ast.ExpressionStatement): The expression statement.

        Returns:
            Any: The expression statement.
        """
        if (
            isinstance(expression_statement.expression, ast.FunctionCall)
            and expression_statement.expression.name.name == "capture_v0"
            and self._register_identifier
        ):
            # For capture_v0 nodes, it replaces it with classical assignment statements
            # of the form:
            # b[0] = capture_v0(...)
            # b[1] = capture_v0(...)
            new_val = ast.ClassicalAssignment(
                # Ideally should use IndexedIdentifier here, but this works since it is just
                # for printing.
                ast.Identifier(name=f"{self._register_identifier}[{self._capture_v0_count}]"),
                ast.AssignmentOperator["="],
                expression_statement.expression,
            )
            self._capture_v0_count += 1
            return new_val
        else:
            return expression_statement



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import pytest

from braket.registers import Qubit, QubitSet


@pytest.fixture
def qubits():
    return QubitSet([0, Qubit(1)])


def test_qubit_set_is_ordered_set():
    qubits = QubitSet([1, 2, 3, 1, 2, 3])
    assert qubits == QubitSet([1, 2, 3])


def test_default_input():
    assert QubitSet() == QubitSet([])


def test_with_single():
    assert QubitSet(0) == tuple([Qubit(0)])


def test_with_iterable():
    assert QubitSet([0, 1]) == tuple([Qubit(0), Qubit(1)])


def test_with_nested_iterable():
    assert QubitSet([0, 1, [2, 3]]) == tuple([Qubit(0), Qubit(1), Qubit(2), Qubit(3)])


def test_with_qubit_set():
    qubits = QubitSet([0, 1])
    assert QubitSet([qubits, [2, 3]]) == tuple([Qubit(0), Qubit(1), Qubit(2), Qubit(3)])


def test_flattening_does_not_recurse_infinitely():
    with pytest.raises(TypeError):  # str instead of expected int
        QubitSet("kaboom")


def test_map_creates_new_object(qubits):
    mapped_qubits = qubits.map({})
    assert mapped_qubits == qubits
    assert mapped_qubits is not qubits


def test_map_happy_case():
    mapping = {Qubit(0): Qubit(11), Qubit(1): Qubit(5)}
    qubits = QubitSet([0, 1])
    mapped_qubits = qubits.map(mapping)

    assert mapped_qubits == QubitSet([11, 5])


def test_hash(qubits):
    assert hash(qubits) == hash(tuple(qubits))



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import functools
import itertools
import math

import numpy as np
import pytest

from braket.circuits import gates
from braket.circuits.circuit import Circuit
from braket.circuits.observables import I, X, Y, Z
from braket.quantum_information import PauliString

ORDER = ["I", "X", "Y", "Z"]
PAULI_INDEX_MATRICES = {
    0: gates.I().to_matrix(),
    1: gates.X().to_matrix(),
    2: gates.Y().to_matrix(),
    3: gates.Z().to_matrix(),
}
SIGN_MAP = {"+": 1, "-": -1}


@pytest.mark.parametrize(
    "pauli_string, string, phase, observable, obs_with_id",
    [
        ("+XZ", "+XZ", 1, X() @ Z(), X() @ Z()),
        ("-ZXY", "-ZXY", -1, Z() @ X() @ Y(), Z() @ X() @ Y()),
        ("YIX", "+YIX", 1, Y() @ X(), Y() @ I() @ X()),
        (PauliString("-ZYXI"), "-ZYXI", -1, Z() @ Y() @ X(), Z() @ Y() @ X() @ I()),
        ("IIXIIIYI", "+IIXIIIYI", 1, X() @ Y(), I() @ I() @ X() @ I() @ I() @ I() @ Y() @ I()),
    ],
)
def test_happy_case(pauli_string, string, phase, observable, obs_with_id):
    instance = PauliString(pauli_string)
    assert str(instance) == string
    assert instance.phase == phase
    pauli_list = list(str(pauli_string))
    if pauli_list[0] in {"-", "+"}:
        pauli_list.pop(0)
    stripped = "".join(pauli_list)
    assert len(instance) == len(stripped)
    assert len(instance) == instance.qubit_count
    for i in range(len(instance)):
        assert ORDER[instance[i]] == stripped[i]
    assert instance == PauliString(pauli_string)
    assert instance == PauliString(instance)
    assert instance.to_unsigned_observable() == observable
    assert instance.to_unsigned_observable(include_trivial=True) == obs_with_id


@pytest.mark.parametrize(
    "other", ["foo", PauliString("+XYZ"), PauliString("-XI"), PauliString("-XYZI")]
)
def test_not_equal(other):
    assert PauliString("-XYZ") != other


@pytest.mark.xfail(raises=ValueError)
def test_none():
    PauliString(None)


@pytest.mark.xfail(raises=ValueError)
@pytest.mark.parametrize("invalid_string", ["XAY", "-BYZ", "+", "-", "xyz", "", None])
def test_invalid_string(invalid_string):
    PauliString(invalid_string)


@pytest.mark.xfail(raises=TypeError)
def test_invalid_type():
    PauliString(1234)


@pytest.mark.xfail(raises=IndexError)
@pytest.mark.parametrize("string", ["XZY", "-YIIXZ", "+IXIYIZ"])
def test_index_out_of_bounds(string):
    PauliString(string)[len(string)]


@pytest.mark.parametrize(
    "string,weight",
    # Make phase explicit for test simplicity
    list(itertools.product(["-ZYX", "-IXIIXYZ", "+ZXYXY"], [1, 2, 3])),
)
def test_weight_n_substrings(string, weight):
    pauli_string = PauliString(string)
    qubit_count = pauli_string.qubit_count
    nontrivial = [qubit for qubit in range(qubit_count) if pauli_string[qubit]]
    substrings = []
    for indices in itertools.combinations(nontrivial, weight):
        factors = [string[qubit + 1] if qubit in indices else "I" for qubit in range(qubit_count)]
        substrings.append(PauliString(f"{string[0]}{''.join(factors)}"))
    actual = pauli_string.weight_n_substrings(weight)
    assert actual == tuple(substrings)
    assert len(actual) == math.comb(len(nontrivial), weight)


@pytest.mark.parametrize(
    "string,signs",
    list(itertools.product(["X", "Y", "Z"], ["+", "-"]))
    + [("ZIY", "+--"), ("YIXIZ", [1, 1, -1, -1, 1]), ("XYZ", (-1, -1, -1)), ("XZIY", None)],
)
def test_eigenstate(string, signs):
    pauli_string = PauliString(string)
    circuit = pauli_string.eigenstate(signs)
    for qubit in range(len(string)):
        circuit.i(qubit)
    initial_state = np.zeros(2 ** len(pauli_string))
    initial_state[0] = 1
    state = circuit.to_unitary() @ initial_state

    if not signs:
        signs = [1] * len(string)
    signs_list = [SIGN_MAP[sign] for sign in signs] if isinstance(signs, str) else signs
    positive = [signs_list[i] for i in range(len(string)) if signs_list[i] < 0 and string[i] != "I"]
    total_sign = 1 if len(positive) % 2 == 0 else -1
    pauli_matrix = functools.reduce(
        np.kron, [PAULI_INDEX_MATRICES[pauli] for pauli in pauli_string]
    )
    actual = total_sign * pauli_matrix @ state
    assert np.allclose(actual, state)


@pytest.mark.xfail(raises=ValueError)
@pytest.mark.parametrize("sign", ["+ab", "--+-", [+2, -1, -1], (-1, 1j, 1)])
def test_eigenstate_invalid_signs(sign):
    PauliString("XYZ").eigenstate(sign)


@pytest.mark.parametrize(
    "circ_arg_1, circ_arg_2, circ_res",
    [
        ("III", "+III", "III"),
        ("Z", "I", "Z"),
        ("I", "-Y", "-Y"),
        ("XYXZY", "+XYXZY", "IIIII"),
        ("XYZ", "ZYX", "YIY"),
        ("YZ", "ZX", "-XY"),
        ("-Z", "Y", "X"),
        ("Z", "Y", "-X"),
    ],
)
def test_dot(circ_arg_1, circ_arg_2, circ_res):
    circ1 = PauliString(circ_arg_1)
    circ2 = circ1.dot(PauliString(circ_arg_2))
    assert circ2 == PauliString(circ_res)
    assert circ1 == circ1

    # Test in-place computation
    circ1 = PauliString(circ_arg_1)
    circ1.dot(PauliString(circ_arg_2), inplace=True)
    assert circ1 == PauliString(circ_res)

    # Test operator overloads
    circ1 = PauliString(circ_arg_1)
    circ2 = PauliString(circ_arg_2)
    assert circ1 * circ2 == PauliString(circ_res)
    circ1 *= circ2
    assert circ1 == PauliString(circ_res)


@pytest.mark.xfail(raises=ValueError)
@pytest.mark.parametrize(
    "circ1, circ2, operation",
    [
        (PauliString("III"), PauliString("II"), "dot()"),
        (PauliString("III"), PauliString("II"), "*"),
        (PauliString("III"), PauliString("II"), "*="),
        (PauliString("IXI"), PauliString("II"), "dot()"),
        (PauliString("IXI"), PauliString("II"), "*"),
        (PauliString("IXI"), PauliString("II"), "*="),
    ],
)
def test_dot_unequal_lengths(circ1, circ2, operation):
    if operation == "dot()":
        circ1.dot(circ2)
    elif operation == "*":
        circ1 * circ2
    elif operation == "*=":
        circ1 *= circ2


@pytest.mark.parametrize(
    "circ, n, circ_res",
    [
        ("-X", 1, "-X"),
        ("Y", 2, "I"),
        ("Y", -2, "I"),
        ("-X", 3, "-X"),
        ("XYZ", 5, "XYZ"),
        ("XYZ", -5, "XYZ"),
        ("XYZ", 6, "III"),
        ("XYZ", 1, "XYZ"),
        ("-YX", 0, "II"),
        ("Y", 0, "I"),
    ],
)
def test_power(circ, n, circ_res):
    circ1 = PauliString(circ)
    circ2 = circ1.power(n)
    assert circ2 == PauliString(circ_res)
    assert circ1 == PauliString(circ)

    # Test in-place computation
    circ1.power(n, inplace=True)
    assert circ1 == PauliString(circ_res)

    # Test operator overloads
    circ1 = PauliString(circ)
    assert (circ1**n) == PauliString(circ_res)
    circ1 **= n
    assert circ1 == PauliString(circ_res)


@pytest.mark.xfail(raises=ValueError)
@pytest.mark.parametrize(
    "circ, n, operation",
    [
        (PauliString("XYZ"), "-1.2", "power()"),
        (PauliString("XYZ"), "-1.2", "**"),
        (PauliString("XYZ"), "-1.2", "**="),
        (PauliString("ZYX"), "1.2", "power()"),
        (PauliString("ZYX"), "1.2", "**"),
        (PauliString("ZYX"), "1.2", "**="),
        (PauliString("I"), "3j", "power()"),
        (PauliString("I"), "3j", "**"),
        (PauliString("I"), "3j", "**="),
    ],
)
def test_power_invalid_exp(circ, n, operation):
    if operation == "power()":
        circ.power(n)
    elif operation == "**":
        circ**n
    elif operation == "**=":
        circ **= n


@pytest.mark.parametrize(
    "circ, circ_res",
    [
        (PauliString("I"), Circuit().i(0)),
        (PauliString("-X"), Circuit().x(0)),
        (PauliString("IYX"), Circuit().i(0).y(1).x(2)),
        (PauliString("ZIIIX"), Circuit().z(0).i(1).i(2).i(3).x(4)),
    ],
)
def test_to_circuit(circ, circ_res):
    assert circ.to_circuit() == circ_res



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from setuptools import find_namespace_packages, setup

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("src/braket/_sdk/_version.py") as f:
    version = f.readlines()[-1].split()[-1].strip("\"'")

setup(
    name="amazon-braket-sdk",
    version=version,
    license="Apache License 2.0",
    python_requires=">= 3.9",
    packages=find_namespace_packages(where="src", exclude=("test",)),
    package_dir={"": "src"},
    install_requires=[
        "amazon-braket-schemas>=1.21.0",
        "amazon-braket-default-simulator>=1.21.2",
        "oqpy~=0.3.5",
        "setuptools",
        "backoff",
        "boltons",
        "boto3>=1.28.53",
        "cloudpickle==2.2.1",
        "nest-asyncio",
        "networkx",
        "numpy<2",
        "openpulse",
        "openqasm3",
        "sympy",
    ],
    extras_require={
        "test": [
            "black",
            "botocore",
            "flake8<=5.0.4",
            "isort",
            "jsonschema==3.2.0",
            "pre-commit",
            "pylint",
            "pytest",
            "pytest-cov",
            "pytest-rerunfailures",
            "pytest-xdist[psutil]",
            "sphinx",
            "sphinx-rtd-theme",
            "sphinxcontrib-apidoc",
            "tox",
        ]
    },
    include_package_data=True,
    url="https://github.com/amazon-braket/amazon-braket-sdk-python",
    author="Amazon Web Services",
    description=(
        "An open source library for interacting with quantum computing devices on Amazon Braket"
    ),
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords="Amazon AWS Quantum",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "Natural Language :: English",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
)



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from __future__ import annotations

from typing import Any, Optional

from braket.circuits.basis_state import BasisState, BasisStateInput
from braket.circuits.compiler_directive import CompilerDirective
from braket.circuits.gate import Gate
from braket.circuits.operator import Operator
from braket.circuits.quantum_operator import QuantumOperator
from braket.circuits.serialization import IRType, SerializationProperties
from braket.registers.qubit import QubitInput
from braket.registers.qubit_set import QubitSet, QubitSetInput

# InstructionOperator is a type alias, and it can be expanded to include other operators
InstructionOperator = Operator


class Instruction:
    """An instruction is a quantum directive that describes the quantum task to perform on a quantum
    device.
    """

    def __init__(
        self,
        operator: InstructionOperator,
        target: Optional[QubitSetInput] = None,
        *,
        control: Optional[QubitSetInput] = None,
        control_state: Optional[BasisStateInput] = None,
        power: float = 1,
    ) -> Instruction:
        """InstructionOperator includes objects of type `Gate` and `Noise` only.

        Args:
            operator (InstructionOperator): Operator for the instruction.
            target (Optional[QubitSetInput]): Target qubits that the operator is applied to.
                Default is None.
            control (Optional[QubitSetInput]): Target qubits that the operator is controlled on.
                Default is None.
            control_state (Optional[BasisStateInput]): Quantum state on which to control the
                operation. Must be a binary sequence of same length as number of qubits in
                `control`. Will be ignored if `control` is not present. May be represented as a
                string, list, or int. For example "0101", [0, 1, 0, 1], 5 all represent
                controlling on qubits 0 and 2 being in the \\|0⟩ state and qubits 1 and 3 being
                in the \\|1⟩ state. Default "1" * len(control).
            power (float): Integer or fractional power to raise the gate to. Negative
                powers will be split into an inverse, accompanied by the positive power.
                Default 1.

        Raises:
            ValueError: If `operator` is empty or any integer in `target` does not meet the `Qubit`
                or `QubitSet` class requirements. Also, if operator qubit count does not equal
                the size of the target qubit set.
            TypeError: If a `Qubit` class can't be constructed from `target` due to an incorrect
                `typing`.

        Examples:
            >>> Instruction(Gate.CNot(), [0, 1])
            Instruction('operator': CNOT, 'target': QubitSet(Qubit(0), Qubit(1)))
            >>> instr = Instruction(Gate.CNot()), QubitSet([0, 1])])
            Instruction('operator': CNOT, 'target': QubitSet(Qubit(0), Qubit(1)))
            >>> instr = Instruction(Gate.H(), 0)
            Instruction('operator': H, 'target': QubitSet(Qubit(0),))
            >>> instr = Instruction(Gate.Rx(0.12), 0)
            Instruction('operator': Rx, 'target': QubitSet(Qubit(0),))
            >>> instr = Instruction(Gate.Rx(0.12, control=1), 0)
            Instruction(
                'operator': Rx,
                'target': QubitSet(Qubit(0),),
                'control': QubitSet(Qubit(1),),
            )
        """
        if not operator:
            raise ValueError("Operator cannot be empty")
        target_set = QubitSet(target)
        control_set = QubitSet(control)
        if isinstance(operator, QuantumOperator) and len(target_set) != operator.qubit_count:
            raise ValueError(
                f"Operator qubit count {operator.qubit_count} must be equal to"
                f" size of target qubit set {target_set}"
            )
        self._operator = operator
        self._target = target_set
        self._control = control_set
        self._control_state = BasisState(
            (1,) * len(control_set) if control_state is None else control_state,
            len(control_set),
        )
        self._power = power

    @property
    def operator(self) -> InstructionOperator:
        """Operator: The operator for the instruction, for example, `Gate`."""
        return self._operator

    @property
    def target(self) -> QubitSet:
        """QubitSet: Target qubits that the operator is applied to."""
        return self._target

    @property
    def control(self) -> QubitSet:
        """QubitSet: Target qubits that the operator is controlled on."""
        return self._control

    @property
    def control_state(self) -> BasisState:
        """BasisState: Quantum state that the operator is controlled to."""
        return self._control_state

    @property
    def power(self) -> float:
        """float: Power that the operator is raised to."""
        return self._power

    def adjoint(self) -> list[Instruction]:
        """Returns a list of Instructions implementing adjoint of this instruction's own operator

        This operation only works on Gate operators and compiler directives.

        Returns:
            list[Instruction]: A list of new instructions that comprise the adjoint of this operator

        Raises:
            NotImplementedError: If `operator` is not of type `Gate` or `CompilerDirective`
        """
        operator = self._operator
        if isinstance(operator, Gate):
            return [
                Instruction(
                    gate,
                    self._target,
                    control=self._control,
                    control_state=self._control_state,
                    power=self._power,
                )
                for gate in operator.adjoint()
            ]
        elif isinstance(operator, CompilerDirective):
            return [Instruction(operator.counterpart(), self._target)]
        raise NotImplementedError(f"Adjoint not supported for {operator}")

    def to_ir(
        self,
        ir_type: IRType = IRType.JAQCD,
        serialization_properties: SerializationProperties | None = None,
    ) -> Any:
        """Converts the operator into the canonical intermediate representation.
        If the operator is passed in a request, this method is called before it is passed.

        Args:
            ir_type(IRType) : The IRType to use for converting the instruction object to its
                IR representation.
            serialization_properties (SerializationProperties | None): The serialization properties
                to use while serializing the object to the IR representation. The serialization
                properties supplied must correspond to the supplied `ir_type`. Defaults to None.

        Returns:
            Any: IR object of the instruction.
        """
        kwargs = {}
        if self.control:
            kwargs["control"] = self.control
            kwargs["control_state"] = self.control_state
        if self.power != 1:
            kwargs["power"] = self.power
        return self._operator.to_ir(
            [int(qubit) for qubit in self._target],
            ir_type=ir_type,
            serialization_properties=serialization_properties,
            **kwargs,
        )

    @property
    def ascii_symbols(self) -> tuple[str, ...]:
        """tuple[str, ...]: Returns the ascii symbols for the instruction's operator."""
        return self._operator.ascii_symbols

    def copy(
        self,
        target_mapping: Optional[dict[QubitInput, QubitInput]] = None,
        target: Optional[QubitSetInput] = None,
        control_mapping: Optional[dict[QubitInput, QubitInput]] = None,
        control: Optional[QubitSetInput] = None,
        control_state: Optional[BasisStateInput] = None,
        power: float = 1,
    ) -> Instruction:
        """Return a shallow copy of the instruction.

        Note:
            If `target_mapping` is specified, then `self.target` is mapped to the specified
            qubits. This is useful apply an instruction to a circuit and change the target qubits.
            Same relationship holds for `control_mapping`.

        Args:
            target_mapping (Optional[dict[QubitInput, QubitInput]]): A dictionary of
                qubit mappings to apply to the target. Key is the qubit in this `target` and the
                value is what the key is changed to. Default = `None`.
            target (Optional[QubitSetInput]): Target qubits for the new instruction.
                Default is None.
            control_mapping (Optional[dict[QubitInput, QubitInput]]): A dictionary of
                qubit mappings to apply to the control. Key is the qubit in this `control` and the
                value is what the key is changed to. Default = `None`.
            control (Optional[QubitSetInput]): Control qubits for the new instruction.
                Default is None.
            control_state (Optional[BasisStateInput]): Quantum state on which to control the
                operation. Must be a binary sequence of same length as number of qubits in
                `control`. Will be ignored if `control` is not present. May be represented as a
                string, list, or int. For example "0101", [0, 1, 0, 1], 5 all represent
                controlling on qubits 0 and 2 being in the \\|0⟩ state and qubits 1 and 3 being
                in the \\|1⟩ state. Default "1" * len(control).
            power (float): Integer or fractional power to raise the gate to. Negative
                powers will be split into an inverse, accompanied by the positive power.
                Default 1.

        Returns:
            Instruction: A shallow copy of the instruction.

        Raises:
            TypeError: If both `target_mapping` and `target` are supplied.

        Examples:
            >>> instr = Instruction(Gate.H(), 0)
            >>> new_instr = instr.copy()
            >>> new_instr.target
            QubitSet(Qubit(0))
            >>> new_instr = instr.copy(target_mapping={0: 5})
            >>> new_instr.target
            QubitSet(Qubit(5))
            >>> new_instr = instr.copy(target=[5])
            >>> new_instr.target
            QubitSet(Qubit(5))
        """
        if target_mapping and target is not None:
            raise TypeError("Only 'target_mapping' or 'target' can be supplied, but not both.")
        if control_mapping and control is not None:
            raise TypeError("Only 'control_mapping' or 'control' can be supplied, but not both.")

        new_target = self._target.map(target_mapping or {}) if target is None else target
        new_control = self._control.map(control_mapping or {}) if control is None else control
        new_control_state = self._control_state if control_state is None else control_state

        return Instruction(
            self._operator,
            new_target,
            control=new_control,
            control_state=new_control_state,
            power=power,
        )

    def __repr__(self):
        return (
            f"Instruction('operator': {self._operator}, "
            f"'target': {self._target}, "
            f"'control': {self._control}, "
            f"'control_state': {self._control_state.as_tuple}, "
            f"'power': {self.power})"
        )

    def __eq__(self, other: Instruction):
        if isinstance(other, Instruction):
            return (
                self._operator,
                self._target,
                self._control,
                self._control_state,
                self._power,
            ) == (
                other._operator,
                other._target,
                other._control,
                other._control_state,
                other._power,
            )
        return NotImplemented

    def __pow__(self, power: float, modulo: float = None):
        new_power = self.power * power
        if modulo is not None:
            new_power %= modulo
        return self.copy(power=new_power)



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from unittest.mock import Mock

import pytest

from braket.ahs.hamiltonian import Hamiltonian
from braket.ahs.shifting_field import ShiftingField
from braket.timings.time_series import StitchBoundaryCondition


@pytest.fixture
def default_shifting_field():
    return ShiftingField(Mock())


def test_create():
    mock0 = Mock()
    field = ShiftingField(magnitude=mock0)
    assert mock0 == field.magnitude


def test_add_hamiltonian(default_shifting_field):
    expected = [default_shifting_field, Mock(), Mock(), Mock()]
    result = expected[0] + Hamiltonian([expected[1], expected[2], expected[3]])
    assert result.terms == expected


def test_add_to_hamiltonian(default_shifting_field):
    expected = [Mock(), Mock(), Mock(), default_shifting_field]
    result = Hamiltonian([expected[0], expected[1], expected[2]]) + expected[3]
    assert result.terms == expected


def test_add_to_other():
    field0 = ShiftingField(Mock())
    field1 = ShiftingField(Mock())
    result = field0 + field1
    assert type(result) is Hamiltonian
    assert result.terms == [field0, field1]


def test_add_to_self(default_shifting_field):
    result = default_shifting_field + default_shifting_field
    assert type(result) is Hamiltonian
    assert result.terms == [default_shifting_field, default_shifting_field]


def test_iadd_to_other(default_shifting_field):
    expected = [Mock(), Mock(), Mock(), default_shifting_field]
    other = Hamiltonian([expected[0], expected[1], expected[2]])
    other += expected[3]
    assert other.terms == expected


def test_from_lists():
    times = [0, 0.1, 0.2, 0.3]
    glob_amplitude = [0.5, 0.8, 0.9, 1.0]
    pattern = [0.3, 0.7, 0.6, -0.5, 0, 1.6]

    sh_field = ShiftingField.from_lists(times, glob_amplitude, pattern)
    assert sh_field.magnitude.time_series.values() == glob_amplitude
    assert sh_field.magnitude.pattern.series == pattern

    assert sh_field.magnitude.time_series.times() == times


@pytest.mark.xfail(raises=ValueError)
def test_from_lists_not_eq_length():
    times = [0, 0.1, 0.2]
    glob_amplitude = [0.5, 0.8, 0.9, 1.0]
    pattern = [0.3, 0.7, 0.6, -0.5, 0, 1.6]

    ShiftingField.from_lists(times, glob_amplitude, pattern)


def test_stitch():
    times_1 = [0, 0.1, 0.2, 0.3]
    glob_amplitude_1 = [0.5, 0.8, 0.9, 1.0]
    pattern_1 = [0.3, 0.7, 0.6, -0.5, 0, 1.6]

    times_2 = [0, 0.1, 0.2, 0.3]
    glob_amplitude_2 = [0.5, 0.8, 0.9, 1.0]
    pattern_2 = pattern_1

    sh_field_1 = ShiftingField.from_lists(times_1, glob_amplitude_1, pattern_1)
    sh_field_2 = ShiftingField.from_lists(times_2, glob_amplitude_2, pattern_2)

    new_sh_field = sh_field_1.stitch(sh_field_2, boundary=StitchBoundaryCondition.LEFT)

    expected_times = [0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6]
    expected_amplitude = glob_amplitude_1 + glob_amplitude_2[1:]
    assert new_sh_field.magnitude.time_series.times() == expected_times
    assert new_sh_field.magnitude.time_series.values() == expected_amplitude
    assert new_sh_field.magnitude.pattern == sh_field_1.magnitude.pattern


@pytest.mark.xfail(raises=ValueError)
def test_stitch_not_eq_pattern():
    times_1 = [0, 0.1, 0.2, 0.3]
    glob_amplitude_1 = [0.5, 0.8, 0.9, 1.0]
    pattern_1 = [0.3, 0.7, 0.6, -0.5, 0, 1.6]

    times_2 = [0.4, 0.5, 0.6, 0.7]
    glob_amplitude_2 = [0.5, 0.8, 0.9, 1.0]
    pattern_2 = [-0.3, 0.7, 0.6, -0.5, 0, 1.6]

    sh_field_1 = ShiftingField.from_lists(times_1, glob_amplitude_1, pattern_1)
    sh_field_2 = ShiftingField.from_lists(times_2, glob_amplitude_2, pattern_2)

    sh_field_1.stitch(sh_field_2)


def test_discretize():
    magnitude_mock = Mock()
    mock_properties = Mock()
    field = ShiftingField(magnitude=magnitude_mock)
    discretized_field = field.discretize(mock_properties)
    magnitude_mock.discretize.assert_called_with(
        time_resolution=mock_properties.rydberg.rydbergLocal.timeResolution,
        value_resolution=mock_properties.rydberg.rydbergLocal.commonDetuningResolution,
        pattern_resolution=mock_properties.rydberg.rydbergLocal.localDetuningResolution,
    )
    assert field is not discretized_field
    assert discretized_field.magnitude == magnitude_mock.discretize.return_value


@pytest.mark.xfail(raises=ValueError)
def test_iadd_to_itself(default_shifting_field):
    default_shifting_field += Hamiltonian(Mock())



from unittest.mock import patch

import pytest

from braket.jobs.metrics import log_metric


@pytest.mark.parametrize(
    "test_value, test_timestamp, test_iteration, result_string",
    [
        # Happy case
        (0.1, 1, 2, "Metrics - timestamp=1; TestName=0.1; iteration_number=2;"),
        # We handle exponent values
        (3.14e-22, 1, 2, "Metrics - timestamp=1; TestName=3.14e-22; iteration_number=2;"),
        # When iteration number is not provided, we don't print it
        (5, 1, None, "Metrics - timestamp=1; TestName=5;"),
        # When iteration number is 0, we do print it
        (5, 1, 0, "Metrics - timestamp=1; TestName=5; iteration_number=0;"),
        # When timestamp is not provided, we use time.time()
        (-3.14, None, 2, "Metrics - timestamp=time_mocked; TestName=-3.14; iteration_number=2;"),
    ],
)
@patch("time.time")
@patch("builtins.print")
def test_log_metric(
    print_mock, time_mock, test_value, test_timestamp, test_iteration, result_string
):
    time_mock.return_value = "time_mocked"
    log_metric(
        metric_name="TestName",
        value=test_value,
        timestamp=test_timestamp,
        iteration_number=test_iteration,
    )
    print_mock.assert_called_with(result_string)



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from __future__ import annotations

import numbers
from typing import Union

QubitInput = Union["Qubit", int]


class Qubit(int):
    """A quantum bit index. The index of this qubit is locally scoped towards the contained
    circuit. This may not be the exact qubit index on the quantum device.
    """

    def __new__(cls, index: int) -> Qubit:
        """Creates a new `Qubit`.

        Args:
            index (int): Index of the qubit.

        Raises:
            ValueError: If `index` is less than zero.

        Returns:
            Qubit: Returns a new Qubit object.

        Examples:
            >>> Qubit(0)
            >>> Qubit(1)
        """
        if not isinstance(index, numbers.Integral):
            raise TypeError(f"Supplied qubit index, {index}, must be an integer.")
        if index < 0:
            raise ValueError(f"Supplied qubit index, {index}, cannot be less than zero.")
        return super().__new__(cls, index)

    def __repr__(self):
        return f"Qubit({super().__repr__()})"

    def __str__(self):
        return self.__repr__()

    @staticmethod
    def new(qubit: QubitInput) -> Qubit:
        """Helper constructor - if input is a `Qubit` it returns the same value,
        else a new `Qubit` is constructed.

        Args:
            qubit (QubitInput): `Qubit` index. If `type == Qubit` then the `qubit` is returned.

        Returns:
            Qubit: The qubit.
        """
        if isinstance(qubit, Qubit):
            return qubit
        else:
            return Qubit(qubit)



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from __future__ import annotations

import copy
import math
from collections.abc import Sequence
from functools import singledispatch
from typing import Optional, Union

from sympy import Float

from braket.circuits.free_parameter_expression import FreeParameterExpression
from braket.circuits.gate import Gate
from braket.circuits.parameterizable import Parameterizable


class AngledGate(Gate, Parameterizable):
    """Class `AngledGate` represents a quantum gate that operates on N qubits and an angle."""

    def __init__(
        self,
        angle: Union[FreeParameterExpression, float],
        qubit_count: Optional[int],
        ascii_symbols: Sequence[str],
    ):
        """Initializes an `AngledGate`.

        Args:
            angle (Union[FreeParameterExpression, float]): The angle of the gate in radians
                or expression representation.
            qubit_count (Optional[int]): The number of qubits that this gate interacts with.
            ascii_symbols (Sequence[str]): ASCII string symbols for the gate. These are used when
                printing a diagram of a circuit. The length must be the same as `qubit_count`, and
                index ordering is expected to correlate with the target ordering on the instruction.
                For instance, if a CNOT instruction has the control qubit on the first index and
                target qubit on the second index, the ASCII symbols should have `["C", "X"]` to
                correlate a symbol with that index.

        Raises:
            ValueError: If the `qubit_count` is less than 1, `ascii_symbols` are `None`, or
                `ascii_symbols` length != `qubit_count`, or `angle` is `None`
        """
        super().__init__(qubit_count=qubit_count, ascii_symbols=ascii_symbols)
        if angle is None:
            raise ValueError("angle must not be None")
        if isinstance(angle, FreeParameterExpression):
            self._parameters = [angle]
        else:
            self._parameters = [float(angle)]  # explicit casting in case angle is e.g. np.float32

    @property
    def parameters(self) -> list[Union[FreeParameterExpression, float]]:
        """Returns the parameters associated with the object, either unbound free parameters or
        bound values.

        Returns:
            list[Union[FreeParameterExpression, float]]: The free parameters or fixed value
            associated with the object.
        """
        return self._parameters

    @property
    def angle(self) -> Union[FreeParameterExpression, float]:
        """Returns the angle of the gate

        Returns:
            Union[FreeParameterExpression, float]: The angle of the gate in radians
        """
        return self._parameters[0]

    def bind_values(self, **kwargs) -> AngledGate:
        """Takes in parameters and attempts to assign them to values.

        Returns:
            AngledGate: A new Gate of the same type with the requested parameters bound.

        Raises:
            NotImplementedError: Subclasses should implement this function.
        """
        raise NotImplementedError

    def adjoint(self) -> list[Gate]:
        """Returns the adjoint of this gate as a singleton list.

        Returns:
            list[Gate]: A list containing the gate with negated angle.
        """
        gate_ascii_name_index = self.ascii_symbols[0].find("(")
        gate_ascii_name = self.ascii_symbols[0][:gate_ascii_name_index]
        new_ascii_symbols = [
            angled_ascii_characters(gate_ascii_name, -self.angle)
        ] * self.qubit_count
        new = copy.copy(self)
        new._parameters = [-angle for angle in self._parameters]
        new._ascii_symbols = new_ascii_symbols
        return [new]

    def __eq__(self, other: AngledGate):
        return (
            isinstance(other, AngledGate)
            and self.name == other.name
            and _angles_equal(self.angle, other.angle)
        )

    def __repr__(self):
        return f"{self.name}('angle': {self.angle}, 'qubit_count': {self.qubit_count})"

    def __hash__(self):
        return hash((self.name, self.angle, self.qubit_count))


class DoubleAngledGate(Gate, Parameterizable):
    """Class `DoubleAngledGate` represents a quantum gate that operates on N qubits and
    two angles.
    """

    def __init__(
        self,
        angle_1: Union[FreeParameterExpression, float],
        angle_2: Union[FreeParameterExpression, float],
        qubit_count: Optional[int],
        ascii_symbols: Sequence[str],
    ):
        """Inits a `DoubleAngledGate`.

        Args:
            angle_1 (Union[FreeParameterExpression, float]): The first angle of the gate in
                radians or expression representation.
            angle_2 (Union[FreeParameterExpression, float]): The second angle of the gate in
                radians or expression representation.
            qubit_count (Optional[int]): The number of qubits that this gate interacts with.
            ascii_symbols (Sequence[str]): ASCII string symbols for the gate. These are used when
                printing a diagram of a circuit. The length must be the same as `qubit_count`, and
                index ordering is expected to correlate with the target ordering on the instruction.
                For instance, if a CNOT instruction has the control qubit on the first index and
                target qubit on the second index, the ASCII symbols should have `["C", "X"]` to
                correlate a symbol with that index.

        Raises:
            ValueError: If `qubit_count` is less than 1, `ascii_symbols` are `None`, or
                `ascii_symbols` length != `qubit_count`, or `angle_1` or `angle_2` is `None`
        """
        super().__init__(qubit_count=qubit_count, ascii_symbols=ascii_symbols)
        if angle_1 is None or angle_2 is None:
            raise ValueError("angles must not be None")
        self._parameters = [
            (
                angle
                if isinstance(angle, FreeParameterExpression)
                else float(angle)  # explicit casting in case angle is e.g. np.float32
            )
            for angle in (angle_1, angle_2)
        ]

    @property
    def parameters(self) -> list[Union[FreeParameterExpression, float]]:
        """Returns the parameters associated with the object, either unbound free parameters or
        bound values.

        Returns:
            list[Union[FreeParameterExpression, float]]: The free parameters or fixed value
            associated with the object.
        """
        return self._parameters

    @property
    def angle_1(self) -> Union[FreeParameterExpression, float]:
        """Returns the first angle of the gate

        Returns:
            Union[FreeParameterExpression, float]: The first angle of the gate in radians
        """
        return self._parameters[0]

    @property
    def angle_2(self) -> Union[FreeParameterExpression, float]:
        """Returns the second angle of the gate

        Returns:
            Union[FreeParameterExpression, float]: The second angle of the gate in radians
        """
        return self._parameters[1]

    def bind_values(self, **kwargs: FreeParameterExpression | str) -> AngledGate:
        """Takes in parameters and attempts to assign them to values.

        Args:
            **kwargs (FreeParameterExpression | str): The parameters that are being assigned.

        Returns:
            AngledGate: A new Gate of the same type with the requested parameters bound.

        Raises:
            NotImplementedError: Subclasses should implement this function.
        """
        raise NotImplementedError

    def adjoint(self) -> list[Gate]:
        """Returns the adjoint of this gate as a singleton list.

        Returns:
            list[Gate]: A list containing the gate with negated angle.
        """
        raise NotImplementedError

    def __eq__(self, other: DoubleAngledGate):
        return (
            isinstance(other, DoubleAngledGate)
            and self.name == other.name
            and _angles_equal(self.angle_1, other.angle_1)
            and _angles_equal(self.angle_2, other.angle_2)
        )

    def __repr__(self):
        return (
            f"{self.name}('angles': ({self.angle_1}, {self.angle_2}), "
            f"'qubit_count': {self.qubit_count})"
        )

    def __hash__(self):
        return hash((self.name, self.angle_1, self.angle_2, self.qubit_count))


class TripleAngledGate(Gate, Parameterizable):
    """Class `TripleAngledGate` represents a quantum gate that operates on N qubits and
    three angles.
    """

    def __init__(
        self,
        angle_1: Union[FreeParameterExpression, float],
        angle_2: Union[FreeParameterExpression, float],
        angle_3: Union[FreeParameterExpression, float],
        qubit_count: Optional[int],
        ascii_symbols: Sequence[str],
    ):
        """Inits a `TripleAngledGate`.

        Args:
            angle_1 (Union[FreeParameterExpression, float]): The first angle of the gate in
                radians or expression representation.
            angle_2 (Union[FreeParameterExpression, float]): The second angle of the gate in
                radians or expression representation.
            angle_3 (Union[FreeParameterExpression, float]): The third angle of the gate in
                radians or expression representation.
            qubit_count (Optional[int]): The number of qubits that this gate interacts with.
            ascii_symbols (Sequence[str]): ASCII string symbols for the gate. These are used when
                printing a diagram of a circuit. The length must be the same as `qubit_count`, and
                index ordering is expected to correlate with the target ordering on the instruction.
                For instance, if a CNOT instruction has the control qubit on the first index and
                target qubit on the second index, the ASCII symbols should have `["C", "X"]` to
                correlate a symbol with that index.

        Raises:
            ValueError: If `qubit_count` is less than 1, `ascii_symbols` are `None`, or
                `ascii_symbols` length != `qubit_count`, or `angle_1` or `angle_2` or `angle_3`
                 is `None`
        """
        super().__init__(qubit_count=qubit_count, ascii_symbols=ascii_symbols)
        if angle_1 is None or angle_2 is None or angle_3 is None:
            raise ValueError("angles must not be None")
        self._parameters = [
            (
                angle
                if isinstance(angle, FreeParameterExpression)
                else float(angle)  # explicit casting in case angle is e.g. np.float32
            )
            for angle in (angle_1, angle_2, angle_3)
        ]

    @property
    def parameters(self) -> list[Union[FreeParameterExpression, float]]:
        """Returns the parameters associated with the object, either unbound free parameters or
        bound values.

        Returns:
            list[Union[FreeParameterExpression, float]]: The free parameters or fixed value
            associated with the object.
        """
        return self._parameters

    @property
    def angle_1(self) -> Union[FreeParameterExpression, float]:
        """Returns the first angle of the gate

        Returns:
            Union[FreeParameterExpression, float]: The first angle of the gate in radians
        """
        return self._parameters[0]

    @property
    def angle_2(self) -> Union[FreeParameterExpression, float]:
        """Returns the second angle of the gate

        Returns:
            Union[FreeParameterExpression, float]: The second angle of the gate in radians
        """
        return self._parameters[1]

    @property
    def angle_3(self) -> Union[FreeParameterExpression, float]:
        """Returns the third angle of the gate

        Returns:
            Union[FreeParameterExpression, float]: The third angle of the gate in radians
        """
        return self._parameters[2]

    def bind_values(self, **kwargs: FreeParameterExpression | str) -> AngledGate:
        """Takes in parameters and attempts to assign them to values.

        Args:
            **kwargs (FreeParameterExpression | str): The parameters that are being assigned.

        Returns:
            AngledGate: A new Gate of the same type with the requested parameters bound.

        Raises:
            NotImplementedError: Subclasses should implement this function.
        """
        raise NotImplementedError

    def adjoint(self) -> list[Gate]:
        """Returns the adjoint of this gate as a singleton list.

        Returns:
            list[Gate]: A list containing the gate with negated angle.
        """
        raise NotImplementedError

    def __eq__(self, other: TripleAngledGate):
        return (
            isinstance(other, TripleAngledGate)
            and self.name == other.name
            and _angles_equal(self.angle_1, other.angle_1)
            and _angles_equal(self.angle_2, other.angle_2)
            and _angles_equal(self.angle_3, other.angle_3)
        )

    def __repr__(self):
        return (
            f"{self.name}('angles': ({self.angle_1}, {self.angle_2}, {self.angle_3}), "
            f"'qubit_count': {self.qubit_count})"
        )

    def __hash__(self):
        return hash((self.name, self.angle_1, self.angle_2, self.angle_3, self.qubit_count))


@singledispatch
def _angles_equal(
    angle_1: Union[FreeParameterExpression, float], angle_2: Union[FreeParameterExpression, float]
) -> bool:
    return isinstance(angle_2, float) and math.isclose(angle_1, angle_2)


@_angles_equal.register
def _(angle_1: FreeParameterExpression, angle_2: FreeParameterExpression):
    return angle_1 == angle_2


def angled_ascii_characters(gate: str, angle: Union[FreeParameterExpression, float]) -> str:
    """Generates a formatted ascii representation of an angled gate.

    Args:
        gate (str): The name of the gate.
        angle (Union[FreeParameterExpression, float]): The angle for the gate.

    Returns:
        str: Returns the ascii representation for an angled gate.

    """
    return f'{gate}({angle:{".2f" if isinstance(angle, (float, Float)) else ""}})'


def _multi_angled_ascii_characters(
    gate: str,
    *angles: Union[FreeParameterExpression, float],
) -> str:
    """Generates a formatted ascii representation of an angled gate.

    Args:
        gate (str): The name of the gate.
        *angles (Union[FreeParameterExpression, float]): angles in radians.

    Returns:
        str: Returns the ascii representation for an angled gate.

    """

    def format_string(angle: FreeParameterExpression | float) -> str:
        """Formats an angle for ASCII representation.

        Args:
            angle (FreeParameterExpression | float): The angle to format.

        Returns:
            str: The ASCII representation of the angle.
        """
        return ".2f" if isinstance(angle, (float, Float)) else ""

    return f"{gate}({', '.join(f'{angle:{format_string(angle)}}' for angle in angles)})"


def get_angle(gate: AngledGate, **kwargs: FreeParameterExpression | str) -> AngledGate:
    """Gets the angle with all values substituted in that are requested.

    Args:
        gate (AngledGate): The subclass of AngledGate for which the angle is being obtained.
        **kwargs (FreeParameterExpression | str): The named parameters that are being filled
            for a particular gate.

    Returns:
        AngledGate: A new gate of the type of the AngledGate originally used with all
        angles updated.
    """
    new_angle = (
        gate.angle.subs(kwargs) if isinstance(gate.angle, FreeParameterExpression) else gate.angle
    )
    return type(gate)(angle=new_angle)


def _get_angles(
    gate: TripleAngledGate, **kwargs: FreeParameterExpression | str
) -> TripleAngledGate:
    """Gets the angle with all values substituted in that are requested.

    Args:
        gate (TripleAngledGate): The subclass of TripleAngledGate for which the angle is being
            obtained.
        **kwargs (FreeParameterExpression | str): The named parameters that are being filled
            for a particular gate.

    Returns:
        TripleAngledGate: A new gate of the type of the AngledGate originally used with all angles
        updated.
    """
    new_angles = [
        (
            getattr(gate, angle).subs(kwargs)
            if isinstance(getattr(gate, angle), FreeParameterExpression)
            else getattr(gate, angle)
        )
        for angle in ("angle_1", "angle_2", "angle_3")
    ]
    return type(gate)(angle_1=new_angles[0], angle_2=new_angles[1], angle_3=new_angles[2])
